#include "Tools.h"
#include "matrix.h"
#include <sstream>
#include <fstream>
#include "minmax.h"
#include <string>
#include <list>
#include <algorithm>
#include <cstdlib>
#include <cmath>

using namespace std ;

/********************/
/*                  */
/* Klasse OrtDichte */
/*                  */
/********************/

class OrtDichte
{ Matrix<float> Ort, Dichte ;
  float Isotropie, Anisotropie ;

  public:
  OrtDichte () 
  : Ort (Null,1,3), Dichte (Null,3,3), Isotropie (0), Anisotropie (0)
  {
  }

  void CheckGrenzen (float& MinIso, float& MaxIso, float& MinAniso, float& MaxAniso, 
       float& MinX, float& MaxX, float& MinY, float& MaxY, float& MinZ, float& MaxZ) ;
  bool Innerhalb (int Ebene, float von, float bis) 
  { return ((*Ort[Ebene] >= von) && (*Ort[Ebene] <= bis)) ; }
  float GetKoordinate (int Ebene) 
  { return (*Ort[Ebene]) ; }
  float GetIsotropie () 
  { return (Isotropie) ; }
  float GetAnisotropie () 
  { return (Anisotropie) ; }
  Matrix<float> GetOrt () 
  { return (Ort) ; } ;
  void write (ostream& OutStream) ;
  friend istream& operator >> (istream& InStream, OrtDichte& OD) ;
  friend ostream& operator << (ostream& OutStream, const OrtDichte& OD) ;
  bool operator == (const OrtDichte& OD) 
  { return (Ort == OD.Ort && Dichte == OD.Dichte) ;
  }
  bool operator < (const OrtDichte& OD) 
  { return (*Ort[2] < *OD.Ort[2]) ;
  }
} ;

void OrtDichte::write (ostream& OutStream) 
{ Ort.write (OutStream) ;
  Dichte.write (OutStream) ;
}

istream& operator >> (istream& InStream, OrtDichte& OD) 
{ InStream >> OD.Ort >> OD.Dichte ;
  OD.Dichte = OD.Dichte.ZeilenSpaltenVertauschen () ;
  OD.Isotropie = OD.Dichte.Isotropie () ;
  OD.Anisotropie = OD.Dichte.Anisotropie () ;
  return (InStream) ;
}

ostream& operator << (ostream& OutStream, const OrtDichte& OD) 
{ OutStream << OD.Ort << endl << OD.Dichte ;
  return (OutStream) ;
}

void OrtDichte::CheckGrenzen (float& MinIso, float& MaxIso, float& MinAniso, float& MaxAniso, 
     float& MinX, float& MaxX, float& MinY, float& MaxY, float& MinZ, float& MaxZ) 
{ if (Isotropie < MinIso) MinIso = Isotropie ;
  if (Isotropie > MaxIso) MaxIso = Isotropie ;
  if (Anisotropie < MinAniso) MinAniso = Anisotropie ;
  if (Anisotropie > MaxAniso) MaxAniso = Anisotropie ;
  if (*Ort[0] < MinX) MinX = *Ort[0] ;
  if (*Ort[0] > MaxX) MaxX = *Ort[0] ;
  if (*Ort[1] < MinY) MinY = *Ort[1] ;
  if (*Ort[1] > MaxY) MaxY = *Ort[1] ;
  if (*Ort[2] < MinZ) MinZ = *Ort[2] ;
  if (*Ort[2] > MaxZ) MaxZ = *Ort[2] ;
}

/******************/
/*                */
/* Klasse Minimax */
/*                */
/******************/

const float Gross = float (1e38) ;

class Minimax
{ private:
  float MinIso,MaxIso,MinAniso,MaxAniso ;
  float MinX,MaxX,MinY,MaxY,MinZ,MaxZ ;

  public:
  Minimax () 
  : MinIso (Gross), MaxIso (-Gross), MinAniso (Gross), MaxAniso (-Gross),
    MinX (Gross), MaxX (-Gross), MinY (Gross), MaxY (-Gross), MinZ (Gross), MaxZ (-Gross)
  {
  }
  
  void operator () (OrtDichte OD) 
  { OD.CheckGrenzen (MinIso,MaxIso,MinAniso,MaxAniso,MinX,MaxX,MinY,MaxY,MinZ,MaxZ) ; 
  }
  float NormOrt (int Ebene,float Wert) ;
  float NormIso (float Wert) ;
  float NormAniso (float Wert) ;
  void Lese (float aMinIso,float aMaxIso,float aMinAniso,float aMaxAniso,float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ) ;
  friend ostream& operator << (ostream& OutStream, const Minimax& MM) ;
} ;

void Minimax::Lese (float aMinIso,float aMaxIso,float aMinAniso,float aMaxAniso,float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ)
{ MinIso = aMinIso ;
  MaxIso = aMaxIso ;
  MinAniso = aMinAniso ;
  MaxAniso = aMaxAniso ;
  MinX = aMinX ;
  MaxX = aMaxX ;
  MinY = aMinY ;
  MaxY = aMaxY ;
  MinZ = aMinZ ;
  MaxZ = aMaxZ ;
}

float Minimax::NormIso (float Wert) 
{ if (MaxIso == MinIso) return (0) ; else 
  { float Ergebnis = (Wert-MinIso)/(MaxIso-MinIso) ;
    if (Ergebnis < 0) Ergebnis = 0 ;
    if (Ergebnis > 1) Ergebnis = 1 ;
    return (Ergebnis) ;
  }
}

float Minimax::NormAniso (float Wert) 
{ if (MaxAniso == MinAniso) return (0) ; else 
  { float Ergebnis = ((Wert)-(MinAniso))/(MaxAniso-MinAniso) ;
    if (Ergebnis < 0) Ergebnis = 0 ;
    if (Ergebnis > 1) Ergebnis = 1 ;
    return (Ergebnis) ;
    //return ((Wert-MinAniso)/(MaxAniso-MinAniso)) ;
  }
}

float Minimax::NormOrt (int Ebene,float Wert) 
{ float Ergebnis ;

  switch (Ebene)
  { case 0: { Ergebnis = (Wert-MinX)/(MaxX-MinX) ; break ; } ;
    case 1: { Ergebnis = (Wert-MinY)/(MaxY-MinY) ; break ; } ;
    case 2: { Ergebnis = (Wert-MinZ)/(MaxZ-MinZ) ; break ; } ;
  }
  if (Ergebnis < 0) Ergebnis = 0 ;
  if (Ergebnis > 1) Ergebnis = 1 ;
  return (Ergebnis) ;  
}

ostream& operator << (ostream& OutStream, const Minimax& MM) 
{ OutStream << MM.MinIso << " <= Isotropie <= " << MM.MaxIso << endl
            << MM.MinAniso << " <= Anisotropie <= " << MM.MaxAniso << endl
            << MM.MinX << " <= X <= " << MM.MaxX << endl
            << MM.MinY << " <= Y <= " << MM.MaxY << endl
            << MM.MinZ << " <= Z <= " << MM.MaxZ << endl ;
  return (OutStream) ;
}

/**********************/
/*                    */
/* globale Funktionen */
/*                    */
/**********************/

void Hilfe () 
{ cout << endl
       << "AICD-extract - Aufbereiten des Outputs von AICD-extract.pl fuer eine" << endl
       << "               nachfolgende Bearbeitung mittels AICD-remap" << endl
       << endl
       << "Dieses Programm wandelt die von AICD-extract.pl extrahierten" << endl
       << "Stromdichtedaten in ein binaeres Format um, das von AICD-remap" << endl
       << "gelesen und weiter verarbeitet werden kann. Die binaere Datei" << endl
       << "erhaelt als Namen den Namen des Output-files von AICD-extract.pl mit" << endl
       << "einem angehaengten \".b\"." << endl
       << endl
       << "Zur Informationen werden die Anzahl der gelesenen Stromdichtetensoren" << endl
       << "und die Minimal- und Maximalwerte der Isotropie der Stromdichte, der" << endl
       << "Anisotropie der Stromdichte und der Ortskoordinaten ausgegeben. Diese" << endl
       << "Informationen koennen durch Aufruf von" << endl 
       << endl
       << "head -12  Binaerdatei.b" << endl 
       << endl
       << "noch einmal ausgegeben werden." << endl
       << endl 
       << "Aufruf: AICD-extract  input-file(s)" << endl
       << "Output: input-file.b" << endl
       << endl ;  
} 

/*****************/
/*               */
/* Hauptprogramm */
/*               */
/*****************/

typedef list<OrtDichte> TStromdichten ;

int main (int argc, char* argv [])
{ char* Inputfilename ;
//  int Ebene ;
//  float von, bis ;

  if (argc == 1) Hilfe () ; else
  for (int ArgIndex = 1; ArgIndex < argc; ++ArgIndex)
  { // Parameter einlesen    
    Inputfilename = argv[ArgIndex] ;
    string Outputfilename = string(Inputfilename)+".b";

    cout << endl << ArgIndex << ". Input-file: " << Inputfilename << endl ;

    // Einlesen der Daten aus dem input-file
    TStromdichten Stromdichte ;
    ifstream Inputfile (Inputfilename) ;

    // Ausgabe in binr formatierte Datei
#ifdef IRIX
    ofstream Bindatei (Outputfilename.c_str()) ;
#else
    ofstream Bindatei (Outputfilename.c_str(), ios::binary) ;
#endif

    OrtDichte Temp;

    while (Inputfile)
    {
      Inputfile >> Temp ;
      if (Inputfile) Stromdichte.push_back (Temp) ;
    }
    cout << "Es wurden " << Stromdichte.size () << " Stromdichtetensoren eingelesen." << endl ;

    // Minimal- und Maximalwerte ermitteln
    Minimax Grenzen (for_each (Stromdichte.begin (), Stromdichte.end (), Minimax ())) ;
    Matrix<float> Mittelpunkt (Null,1,3) ;
    Mittelpunkt = ((*Stromdichte.begin()).GetOrt ()+(*(++Stromdichte.begin())).GetOrt ())/ float (2.0) ;
    float Radius = 0 ;
    {
      for (TStromdichten::iterator Index = Stromdichte.begin(); Index != Stromdichte.end(); ++Index)
          Radius = max(Radius,(Mittelpunkt - (*Index).GetOrt ()).Laenge ()) ;
    }

    cout << endl << "Grenzen:" << endl << Grenzen << endl << "Atomposition:" << endl
         << Mittelpunkt << endl << "Radius: " << Radius << endl ;

    Bindatei << "Anzahl der Stromdichtetensoren: " << Stromdichte.size () << endl
             << "Grenzen:" << endl << Grenzen
             << "Atomposition:" << endl
             << Mittelpunkt << "Radius: " << Radius << endl ;

    // einzelne Tensoren ausgeben
//
// Sortieren darf nicht durchgefhrt werden, weil das vom MSC erstellte Programm dann maximal 32767 Tensoren
// ausgeben kann!!
//	
//    Stromdichte.sort () ;
    {
      for (TStromdichten::iterator Index = Stromdichte.begin(); Index != Stromdichte.end(); ++Index)
          (*Index).write (Bindatei) ;
    }
  }
  return 0;
}

