#!/usr/bin/perl

################
#              #
# AICD-extract #
#              #
################

sub sqr
{
  return ( $_[0] * $_[0] ) ;
}

sub MakePDBFile
{
  my @AtomSymbole =
  ( "Dummy",
    "H","He",
    "Li","Be","B","C","N","O","F","Ne",
    "Na","Mg","Al","Si","P","S","Cl","Ar",
    "K","Ca","Sc","Ti","V","Cr","Mn","Fe","Co","Ni","Cu","Zn","Ga","Ge","As","Se","Br","Kr",
    "Rb","Sr","Y","Zr","Nb","Mo","Tc","Ru","Rh","Pd","Ag","Cd","In","Sn","Sb","Te","I","Xe",
    "Cs","Ba",
    "La","Ce","Pr","Nd","Pm","Sm","Eu","Gd","Tb","Dy","Ho","Er","Tm","Yb","Lu",
    "Hf","Ta","W","Re","Os","Ir","Pt","Au","Hg","Tl","Pb","Bi","Po","At","Rn",
    "Fr","Ra",
    "Ac","Th","Pa","U","Np","Pu","Am","Cm","Bk","Cf","Es","Fm","Md","No","Lr"
  ) ;

  my %AtomRadius =
  ( "H" => 0.230,"He" => 0.700,"Li" => 0.680,"Be" => 0.350,"B" => 0.830,"C" => 0.680,"N" => 0.680,
    "O" => 0.680,"F" => 0.640,"Ne" => 0.700,"Na" => 0.970,"Mg" => 1.100,"Al" => 1.350,"Si" => 1.200,
    "P" => 0.750,"S" => 1.020,"Cl" => 0.990,"Ar" => 0.700,"K" => 1.330,"Ca" => 0.990,"Sc" => 1.440,
    "Ti" => 1.470,"V" => 1.330,"Cr" => 1.350,"Mn" => 1.350,"Fe" => 1.340,"Co" => 1.330,"Ni" => 1.500,
    "Cu" => 1.520,"Zn" => 1.450,"Ga" => 1.220,"Ge" => 1.170,"As" => 1.210,"Se" => 1.220,"Br" => 1.210,
    "Kr" => 1.910,"Rb" => 1.470,"Sr" => 1.120,"Y" => 1.780,"Zr" => 1.560,"Nb" => 1.480,"Mo" => 1.470,
    "Tc" => 1.350,"Ru" => 1.400,"Rh" => 1.450,"Pd" => 1.500,"Ag" => 1.590,"Cd" => 1.690,"In" => 1.630,
    "Sn" => 1.460,"Sb" => 1.460,"Te" => 1.470,"I" => 1.400,"Xe" => 1.980,"Cs" => 1.670,"Ba" => 1.340,
    "La" => 1.870,"Ce" => 1.830,"Pr" => 1.820,"Nd" => 1.810,"Pm" => 1.800,"Sm" => 1.800,"Eu" => 1.990,
    "Gd" => 1.790,"Tb" => 1.760,"Dy" => 1.750,"Ho" => 1.740,"Er" => 1.730,"Tm" => 1.720,"Yb" => 1.940,
    "Lu" => 1.720,"Hf" => 1.570,"Ta" => 1.430,"W" => 1.370,"Re" => 1.350,"Os" => 1.370,"Ir" => 1.320,
    "Pt" => 1.500,"Au" => 1.500,"Hg" => 1.700,"Tl" => 1.550,"Pb" => 1.540,"Bi" => 1.540,"Po" => 1.680,
    "At" => 0.700,"Rn" => 2.400,"Fr" => 2.000,"Ra" => 1.900,"Ac" => 1.880,"Th" => 1.790,"Pa" => 1.610,
    "U" => 1.580,"Np" => 1.550,"Pu" => 1.530,"Am" => 1.510,"Cm" => 0.000,"Bk" => 0.000,"Cf" => 0.000,
    "Es" => 0.000,"Fm" => 0.000,"Md" => 0.000,"No" => 0.000,"Lr" => 0.000
  ) ;
  my $Ergebnis ;
  my $G98out = shift ;
  my $StandardOrientation = shift ;
  
  # g98-output (archive-entry) interpretieren
  # 1. aufteilen nach Leerzeilen
  (undef,undef,$Kommentar) = split /\n\n/,$G98out ;
  
  # 2. Geometrie aufteilen nach einzelnen Zeilen
  $StandardOrientation =~ s/  */ /g ;
  @GeoZeilen = split /\n/,$StandardOrientation ;
  
  # 3. erste Zeile verwerfen, fr alle anderen Zeilen: aufteilen nach Atomsymbol und Koordinaten
  for ($Index = 0; $Index <= $#GeoZeilen; $Index++)
  {
    (undef,undef,$AtomSymbol[$Index],undef,$X[$Index],$Y[$Index],$Z[$Index]) = split / /,$GeoZeilen[$Index] ;
    $AtomSymbol[$Index] = $AtomSymbole[$AtomSymbol[$Index]] ;
  }
  
  # Header schreiben
  $Ergebnis = "HEADER    PROTEIN\nCOMPND    $Kommentar \nAUTHOR    AICD-extract\n" ;

  # Atomkoordinaten schreiben
  for ($Index = 0; $Index <= $#GeoZeilen; $Index++)
  {
    $Ergebnis = $Ergebnis . sprintf "ATOM   %4d  %-4sUNK     1   %9.3f%8.3f%8.3f  1.00  0.00 \n",
                $Index+1,$AtomSymbol[$Index],$X[$Index],$Y[$Index],$Z[$Index] ;
  }
  
  # Bindungen ermitteln
  for ($Index = 0; $Index <= $#GeoZeilen; $Index++)
  {
    # $Connections: reset
    $Connections = "" ;
    
    for ($Index2 = 0; $Index2 <= $#GeoZeilen; $Index2++)
    {
      if (($Index != $Index2) and
          (sqr ($X[$Index]-$X[$Index2]) + sqr ($Y[$Index]-$Y[$Index2]) + sqr ($Z[$Index]-$Z[$Index2])
          < sqr ($AtomRadius{$AtomSymbol[$Index]} + $AtomRadius{$AtomSymbol[$Index2]} + 0.45)))
      {
        $Connections = $Connections . sprintf "%5d",$Index2+1 ;
      }
    }
    
    # Wenn Bindungen gefunden wurden, dann bitte ausgeben
    if (length ($Connections) != 0)
    {
      $Ergebnis = $Ergebnis . sprintf "CONECT%5d$Connections\n",$Index+1 ;
    }
  }
        
  # footer schreiben
  $Ergebnis = $Ergebnis . sprintf "MASTER        0    0    0    0    0    0    0    0 %4d    0 %4d    0\n",$#GeoZeilen+1,$#GeoZeilen+1 ;
  $Ergebnis = $Ergebnis . "END\n" ;
  return ($Ergebnis) ;
}

# Gaussian98-log-file lesen und in die einzelnen jobs aufteilen

sub Extraktion
{
  my $Handle ;
  my $basename ;
  my $state = 0 ;
  my $substate = 1 ;
  my $linecount = 10 ;
  my $job = 1 ;
  my $archiveentry = "" ;
  my @ICDZeilen = () ;
  my $Index = 0 ;
  my $StandardOrientation ;
  
  # Parameter: Dateihandle, basename
  $Handle = shift ;
  $basename = shift ;
  $basename =~ s/\.log$|\.out$// ;
  
  # output-Dateien ffnen
  open (LOGFILE,"> $basename.$job.log") or die ("Die Datei \"$basename.$job.log\" kann nicht zum Schreiben geffnet werden: \n$!") ;
  open (PDBFILE,"> $basename.$job.pdb") or die ("Die Datei \"$basename.$job.pdb\" kann nicht zum Schreiben geffnet werden: \n$!") ;
  open (ICDFILE,"> $basename.$job.icd-raw") or die ("Die Datei \"$basename.$job.icd-raw\" kann nicht zum Schreiben geffnet werden: \n$!") ;
  
  # Alle Zeilen der Reihe nach einlesen
  while (<$Handle>)
  {
    # Zustand 0: nur in LOGFILE schreiben; nach 10 Zeilen Sprung in Zustand 1
    # Dieser Zustand ist nur am Beginn der Datei aktiv und dient dazu
    # die erste "Initial command" Zeile zu berlesen, die einen neuen Job 
    # ankndigt.
    if ($state == 0)
    {
      sub state0 {return 0; }
      # print "\nIn state 0: skimming leading lines.";
      print LOGFILE ;
      if ($linecount-- == 0)
      {
        $state = 1 ;
      }
    }
  
    # Zustand 1: nur in LOGFILE schreiben; Sprung in Zustand 2, 3 und 4 mglich
    elsif ($state == 1)
    {
      sub state1 {return 1; }
      # print "\nState 1: nur in LOGFILE schreiben";
      # Wenn ein neuer job beginnt, dann werden neue output-files erzeugt; Zustand bleibt 1
      if (/^ Initial command:/)
      {
        # print "state 1: Anfang eines neuen Jobs";
        close (ICDFILE) ;
        close (PDBFILE) ;
        close (LOGFILE) ;
        $job++ ;
        open (LOGFILE,"> $basename.$job.log") or die ("Die Datei \"$basename.$job.log\" kann nicht zum Schreiben geffnet werden: \n$!") ;
        open (PDBFILE,"> $basename.$job.pdb") or die ("Die Datei \"$basename.$job.pdb\" kann nicht zum Schreiben geffnet werden: \n$!") ;
        open (ICDFILE,"> $basename.$job.icd-raw") or die ("Die Datei \"$basename.$job.icd-raw\" kann nicht zum Schreiben geffnet werden: \n$!") ;
        print LOGFILE ;
      }

      # Wenn der Bereich mit den Stromdichtedaten anfngt, dann nach Zustand 2 wechseln
      elsif (/^ Gitterpunkte:/)
      {
        # print "Gitterpunkte found";
        $state = 2 ;
        $substate = 1 ;
        $Index = 0 ;
      }
      
      # Wenn der archive-entry anfngt, dann nach Zustand 3 wechseln
      # following line changed by F.K. for compatibility reasons with other
      # gaussian installations
      #     elsif (/^ 1\\1\\GINC-/)
      elsif (/^ 1\\1\\/)
      {
        # print "archive entry found";
        print LOGFILE ;
        $state = 3 ;
        chomp ;
        s/^ // ;
        $archiveentry = $_ ;
      }
      
      # Wenn eine Standard Orientation anfngt, dann nach Zustand 4 wechseln
      elsif (/$Standard_Orientation_Keyword/)
      {
        # print "Standard Orientation found";
        print LOGFILE ;
        $state = 4 ;
        $substate = 4 ;
        $StandardOrientation = "" ;
      }
      
      # Wenn kein Muster erkannt worden ist, wird einfach weiter in LOGFILE geschrieben
      else
      {
        print LOGFILE ;
      }
    }
    
    # Zustand 2: bestimmte Zeilen in LOGFILE schreiben, andere Zeilen in ICDFile schreiben
    # Sprung in Zustand 1 mglich
    elsif ($state == 2)
    {
      sub state2 {return 2; }
      # print "\nIn state 2: separate log / icd";
     
      # Wenn der Bereich mit den Stromdichtedaten aufhrt, dann nach Zustand 1 wechseln
      if (/^     Magnetic properties \(CSGT method\)/)
      {
        # print "Magnetic properties found";
        # Alle gesammelten Zeilen in ICDFILE schreiben
        print ICDFILE @ICDZeilen ;
        @ICDZeilen = () ;
  
        print LOGFILE ;
        $state = 1 ;
      }
      
      # Wenn der Text "Integrating Atom" folgt, dann nur in LOGFILE schreiben
      elsif (/^   Integrating Atom/)
      {
        # print "Integrating Atom found.";
        print LOGFILE ;
      }
      
      # Wenn Gitterpunkte folgt: substate 1
      elsif (/^ Gitterpunkte:/)
      {
        # print "Gitterpunkte in state 2 found";
        # Alle gesammelten Zeilen in ICDFILE schreiben
        print ICDFILE @ICDZeilen ;
        @ICDZeilen = () ;
  
        $substate = 1 ;
        $Index = 0 ;
      }
      
      # Wenn die erste Spalte der Tensoren folgt: substate 2
      elsif (/^ Stromdichtetensor, II=  *1/)
      {
        # print "Stromdichtetensor 1 found";
        $substate = 2 ;
        $Index = 0 ;
      }
      
      # Wenn die zweite Spalte der Tensoren folgt: substate 3
      elsif (/^ Stromdichtetensor, II=  *2/)
      {
        # print "Stromdichtetensor 2 found";
        $substate = 3 ;
        $Index = 0 ;
      }
      
      # Wenn die dritte Spalte der Tensoren folgt: substate 4
      elsif (/^ Stromdichtetensor, II=  *3/)
      {
        # print "Stromdichtetensor 3 found";
        $substate = 4 ;
        $Index = 0 ;
      }
      
      else
      { 
        # D durch E ersetzen in Real Zahl
        s/D/E/g ;
  
        # auf Daten-Muster testen
        # altes Datenmuster rechnet mit 4 Chars vor erster Zahl und 
        # scheitert an neuem g03 Format. Neues Datenmuster parst
        # Leerstellen + unsigned Integer vor den Real Values. (F.K.)
        # if (/^[0-9 ][0-9 ][0-9 ][0-9 ]([- ][-0 ]\.[0-9]*E[+-][0-9]*)([- ][-0 ]\.[0-9]*E[+-][0-9]*)([- ][-0 ]\.[0-9]*E[+-][0-9]*)$/)
        if (/^ *[0-9]*([- ][-0 ]\.[0-9]*E[+-][0-9]*)([- ][-0 ]\.[0-9]*E[+-][0-9]*)([- ][-0 ]\.[0-9]*E[+-][0-9]*)$/)
        {
          if ($substate == 1)
          {
            $ICDZeilen [$Index] = "$1 $2 $3" ;
            $Index++ ;
          }
          elsif ($substate != 4)
          {
            $ICDZeilen [$Index] = $ICDZeilen [$Index] . " $1 $2 $3" ;
            $Index++ ;
          }
          else
          {
            $ICDZeilen [$Index] = $ICDZeilen [$Index] . " $1 $2 $3\n" ;
            $Index++ ;
          }
        }
      }
    }
    
    # Zustand 3: in LOGFILE schreiben und in PDBFile schreiben
    # Sprung in Zustand 1 mglich
    elsif ($state == 3)
    {
      sub state3 {return 3; }
      # print "\nIn state 3: write pdb";
      print LOGFILE ;
      chomp ;
      s/^ // ;
      $archiveentry = $archiveentry . $_ ;

      # Wenn der archive-entry aufhrt, dann nach Zustand 1 wechseln
      if ($archiveentry =~ /\\\@$/)
      {
        $archiveentry =~ s/\\/\n/g ;
        $archiveentry = MakePDBFile ($archiveentry,$StandardOrientation) ;
        print PDBFILE $archiveentry ;
        $state = 1 ;
      }
    }

    # Zustand 4: in LOGFILE schreiben und Standard Orientation merken
    # Sprung in Zustand 1 mglich
    elsif ($state == 4)
    {
      sub state4 {return 4; }
      # print "\nIn state 4: recover standard orientation";
      print LOGFILE ;
      if ($substate > 0)
      {
        # count down substates to read over table header
        $substate-- ;
      }
      else
      {
        # Wenn die standard orientation aufhrt, dann nach Zustand 1 wechseln
        if (/^ ----/)
        {
          $state = 1 ;
        }
        else
        {
          $StandardOrientation = $StandardOrientation . $_ ;
        }
      }
    }
  }
  # print "\n";
  
  # output-Dateien schlieen
  close (ICDFILE) ;
  close (PDBFILE) ;
  close (LOGFILE) ;
}

############
#          #
# Haupteil #
#          #
############
$Standard_Orientation_Keyword = "Standard orientation:";

if (not @ARGV)
{
  print
"
Zweck: Extraktion der induzierten Stromdichte (ICD)
       aus einem oder mehreren Gaussian98-log-files.
       
Dieses Script liest die angegebenen Dateien oder,
falls \"-\" angegeben ist, von STDIN.
die Option -nostd liest die Geometrie aus der ursprnglichen
Z-Matrix (fr IOP=(2/15=1))

Es werden Dateien mit den Extensionen .icd-raw (beinhalten
die Stromdichte-Tensoren) und .pdb (beinhalten die Mole-
klgeometrie) erzeugt. Die Gaussian98-log-files
werden ohne die Stromdichte-Daten neu geschrieben.
      
Aufruf: $0  {-nostd} input-file(s) | -

" ;
}
else
{ 
  ARGUMENT: while (@ARGV)
  {
    # ersten Dateinamen ermitteln
    $Dateiname = shift @ARGV ;
  
    # entweder ffnen oder von STDIN lesen
    if ($Dateiname eq "-")
    {
      # Lesen von STDIN
      $Handle = \*STDIN ;
      $BitteSchliessen = 0 ;
      $Dateiname = "AICD" ;
    }
    elsif ($Dateiname eq "-nostd")
    {
      $Standard_Orientation_Keyword = "Z-Matrix orientation:";
      next ARGUMENT;
    } 
    else
    {
      # Lesen aus den angegebenen Dateien
      open (DATEIHANDLE,$Dateiname) or die ("Die Datei \"$Dateiname\" kann nicht zum Lesen geffnet werden: \n$!") ;
      $Handle = \*DATEIHANDLE ;
      $BitteSchliessen = 1 ;
    }

    # aus Datei "$Dateiname" lesen und weitere Dateien schreiben
    Extraktion ($Handle,$Dateiname) ;
    
    # Datei wieder schlieen
    if ($BitteSchliessen)
    {
      close ($Handle) ;
    }
  }
}

# 0.196588E+01 -0.116970E+01  0.254517E+01 -0.698977E+01 -0.231703E+02 -0.416428E+01  0.329326E+00 -0.602480E+01 -0.700067E+01  0.465065E+01  0.529149E+01  0.607700E+01
# 0.196530E+01 -0.116970E+01  0.254517E+01 -0.698249E+01 -0.231556E+02 -0.414148E+01  0.427348E+00 -0.601475E+01 -0.754198E+01  0.471315E+01  0.574981E+01  0.608376E+01
