#include "Tools.h"
#include "matrix.h"
#include <sstream>
#include <fstream>
#include <iomanip>
#include <unistd.h>
#include "minmax.h"
//#include <process.h>
#include <string>
#include <list>
#include <algorithm>
#include <cstdlib>
#include <cmath>
#include <utility>

using namespace std ;


/******************/
/*                */
/* Klasse Minimax */
/*                */
/******************/

const float Gross = float (1e38) ;

class Minimax
{ private:
  float MinIso,MaxIso,MinAniso,MaxAniso ;
  float MinX,MaxX,MinY,MaxY,MinZ,MaxZ ;

  public:
  Minimax () 
  : MinIso (Gross), MaxIso (-Gross), MinAniso (Gross), MaxAniso (-Gross),
    MinX (Gross), MaxX (-Gross), MinY (Gross), MaxY (-Gross), MinZ (Gross), MaxZ (-Gross)
  {
  }
  
  float NormOrt (int Ebene,float Wert) ;
  float RealOrt (int Ebene,float Wert) ;
  float NormIso (float Wert) ;
  float NormAniso (float Wert) ;
  float DiffX () { return (MaxX-MinX) ; } ;
  float DiffY () { return (MaxY-MinY) ; } ;
  float DiffZ () { return (MaxZ-MinZ) ; } ;
  void CheckGrenzen (const Minimax& MM)
  { CheckGrenzen (MM.MinIso,MM.MaxIso,MM.MinAniso,MM.MaxAniso,MM.MinX,MM.MaxX,MM.MinY,MM.MaxY,MM.MinZ,MM.MaxZ) ;
  }
  void CheckGrenzen (float aMinIso,float aMaxIso,float aMinAniso,float aMaxAniso,float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ) ;
  friend ostream& operator << (ostream& OutStream, const Minimax& MM) ;
  friend istream& operator >> (istream& InStream, Minimax& MM) ;
} ;

void Minimax::CheckGrenzen (float aMinIso,float aMaxIso,float aMinAniso,float aMaxAniso,float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ)
{ if (MinIso > aMinIso) MinIso = aMinIso ;
  if (MaxIso < aMaxIso) MaxIso = aMaxIso ;
  if (MinAniso > aMinAniso) MinAniso = aMinAniso ;
  if (MaxAniso < aMaxAniso) MaxAniso = aMaxAniso ;
  if (MinX > aMinX) MinX = aMinX ;
  if (MaxX < aMaxX) MaxX = aMaxX ;
  if (MinY > aMinY) MinY = aMinY ;
  if (MaxY < aMaxY) MaxY = aMaxY ;
  if (MinZ > aMinZ) MinZ = aMinZ ;
  if (MaxZ < aMaxZ) MaxZ = aMaxZ ;
}

float Minimax::NormIso (float Wert) 
{ if (MaxIso == MinIso) return (0) ; else 
  { float Ergebnis = (Wert-MinIso)/(MaxIso-MinIso) ;
    if (Ergebnis < 0) Ergebnis = 0 ;
    if (Ergebnis > 1) Ergebnis = 1 ;
    return (Ergebnis) ;
  }
}

float Minimax::NormAniso (float Wert) 
{ if (MaxAniso == MinAniso) return (0) ; else 
  { float Ergebnis = ((Wert)-(MinAniso))/(MaxAniso-MinAniso) ;
    if (Ergebnis < 0) Ergebnis = 0 ;
    if (Ergebnis > 1) Ergebnis = 1 ;
    return (Ergebnis) ;
    //return ((Wert-MinAniso)/(MaxAniso-MinAniso)) ;
  }
}

float Minimax::NormOrt (int Ebene,float Wert) 
{ float Ergebnis ;

  switch (Ebene)
  { case 0: { Ergebnis = (Wert-MinX)/(MaxX-MinX) ; break ; } ;
    case 1: { Ergebnis = (Wert-MinY)/(MaxY-MinY) ; break ; } ;
    case 2: { Ergebnis = (Wert-MinZ)/(MaxZ-MinZ) ; break ; } ;
  }
  if (Ergebnis < 0) Ergebnis = 0 ;
  if (Ergebnis > 1) Ergebnis = 1 ;
  return (Ergebnis) ;  
}

float Minimax::RealOrt (int Ebene,float Wert) 
{ float Ergebnis ;
  
  if (Wert < 0) Wert = 0 ;
  if (Wert > 1) Wert = 1 ;

  switch (Ebene)
  { case 0: { Ergebnis = Wert*(MaxX-MinX)+MinX ; break ; } ;
    case 1: { Ergebnis = Wert*(MaxY-MinY)+MinY ; break ; } ;
    case 2: { Ergebnis = Wert*(MaxZ-MinZ)+MinZ ; break ; } ;
  }
  return (Ergebnis) ;
}

ostream& operator << (ostream& OutStream, const Minimax& MM) 
{ OutStream << MM.MinIso << " <= Isotropie <= " << MM.MaxIso << endl
            << MM.MinAniso << " <= Anisotropie <= " << MM.MaxAniso << endl
            << MM.MinX << " <= X <= " << MM.MaxX << endl
            << MM.MinY << " <= Y <= " << MM.MaxY << endl
            << MM.MinZ << " <= Z <= " << MM.MaxZ << endl ;
  return (OutStream) ;
}

istream& operator >> (istream& InStream, Minimax& MM) 
{ string Dummy ;

  InStream >> MM.MinIso >> Dummy >> Dummy >> Dummy >> MM.MaxIso
           >> MM.MinAniso >> Dummy >> Dummy >> Dummy >> MM.MaxAniso
           >> MM.MinX >> Dummy >> Dummy >> Dummy >> MM.MaxX
           >> MM.MinY >> Dummy >> Dummy >> Dummy >> MM.MaxY
           >> MM.MinZ >> Dummy >> Dummy >> Dummy >> MM.MaxZ ;
  return (InStream) ;
}

/*******************/
/*                 */
/* Klasse KartNetz */
/*                 */
/*******************/

class KartNetz
{ private:
  Minimax Grenzen ;
  float* Daten ;
  int AnzahlGitterpunkte ;
  Matrix<int> Max ;

  public:
  KartNetz () : AnzahlGitterpunkte (0), Daten (0), Max (Null,1,3) {} ;
  KartNetz (istream& InStream) ;
  ~KartNetz () ;

  void ParameterAusgabe (ostream& OutStream) const ;
  string ErmittleNeueGrenzen (float AnisotropieLimit) ;

  float* At (int x,int y,int z) ;
  inline float RealWert (int Koordinate, int Wert) ;
} ;

KartNetz::KartNetz (istream& InStream)
: AnzahlGitterpunkte (0), Max (Null,1,3)
{ string Dummy ;

  InStream >> Dummy >> AnzahlGitterpunkte >> Dummy >> *Max[0] >> Dummy >> *Max[1] >> Dummy >> *Max[2] 
           >> Dummy >> Grenzen ;
  InStream.ignore (1) ;

//  cout << "Streampos: " << InStream.tellg () << endl ;

  assert (AnzahlGitterpunkte == Max.x()*Max.y()*Max.z()) ;
  Daten = new float [AnzahlGitterpunkte*9] ;
  InStream.read ((char *) Daten,sizeof (float)*9*AnzahlGitterpunkte) ;

/*
     for (int Index = 0; Index < 10; Index++)
   { for (int Index2 = 0; Index2 < 9; Index2++)
       cout << Daten[Index*9+Index2] << "  " ;
     cout << endl ;
   }
*/
}

KartNetz::~KartNetz ()
{ delete [] Daten ;
}

inline float KartNetz::RealWert (int Koordinate, int Wert)
{ return (Grenzen.RealOrt (Koordinate,float (Wert) / (*Max[Koordinate]-1))) ;
}

float* KartNetz::At (int x,int y,int z)
{ return (&Daten[((z*Max.y()+y)*Max.x()+x)*9]) ;
}

void KartNetz::ParameterAusgabe (ostream& OutStream) const
{ OutStream << AnzahlGitterpunkte << " = " << *Max[0] << " * " << *Max[1] << " * " << *Max[2] << endl
            << "Grenzen:" << endl << Grenzen << endl ;
}

string KartNetz::ErmittleNeueGrenzen (float AnisotropieLimit)
{ bool* Innerhalb = new bool [AnzahlGitterpunkte] ;
  bool* Oberflaeche = new bool [AnzahlGitterpunkte] ;  
  float* Anisotropie = new float [AnzahlGitterpunkte] ;
  int X,Y,Z ;

  // ber Daten iterieren
  bool* InnerhalbAkt = Innerhalb ;
  float* AnisotropieAkt = Anisotropie ;
  for (Z = 0; Z < Max.z(); ++Z)
  { for (Y = 0; Y < Max.y(); ++Y)
    { for (X = 0; X < Max.x(); ++X)
      { // Anisotropie berechnen
        Matrix<float> CurMatrix (At(X,Y,Z),3,3) ;
        *AnisotropieAkt = CurMatrix.Anisotropie () ;
        *InnerhalbAkt++ = *AnisotropieAkt++ >= AnisotropieLimit ;
      }
    }
  }

  cout << "Punkte auf der Isooberflche" << endl ;
  int NeuesStartX = Max.x () ;
  int NeuesStartY = Max.y () ;
  int NeuesStartZ = Max.z () ;
  int NeuesEndX = 0 ;
  int NeuesEndY = 0 ;
  int NeuesEndZ = 0 ;

  // ber Innerhalb iterieren
  InnerhalbAkt = Innerhalb ;
  bool* OberflaecheAkt = Oberflaeche ;
  for (Z = 0; Z < Max.z(); ++Z)
  { cout << "Z = " << Z << endl ;

    int minusZ = Max.y()*Max.x() ;
    if (Z == 0) minusZ = 0 ;
    int plusZ = Max.y()*Max.x() ;
    if (Z == Max.z()-1) plusZ = 0 ;

    for (Y = 0; Y < Max.y(); ++Y)
    { int minusY = Max.x() ;
      if (Y == 0) minusY = 0 ;
      int plusY = Max.x() ;
      if (Y == Max.y()-1) plusY = 0 ;

      for (X = 0; X < Max.x(); ++X)
      { int minusX = 1 ;
        if (X == 0) minusX = 0 ;
        int plusX = 1 ;
        if (X == Max.x()-1) plusX = 0 ;
   
        // Testen, ob der Punkt innerhalb ist und nicht nur Nachbarn hat, die innerhalb sind
        *OberflaecheAkt++ = (*InnerhalbAkt && !(
           *(InnerhalbAkt+plusX) && *(InnerhalbAkt-minusX) && 
           *(InnerhalbAkt+plusY) && *(InnerhalbAkt-minusY) &&
           *(InnerhalbAkt+plusZ) && *(InnerhalbAkt-minusZ) )) ;
        ++InnerhalbAkt ;

        if (*(OberflaecheAkt-1)) 
        { cout << '#' ;

      // Grenzen testen
      NeuesStartX = min (X,NeuesStartX) ;
      NeuesEndX = max (X,NeuesEndX) ;
      NeuesStartY = min (Y,NeuesStartY) ;
      NeuesEndY = max (Y,NeuesEndY) ;
      NeuesStartZ = min (Z,NeuesStartZ) ;
      NeuesEndZ = max (Z,NeuesEndZ) ;
        } else cout << ((*(InnerhalbAkt-1))? 'o' : '.') ;
      }
      cout << endl ;
    }
    cout << endl ;
  }

  // Grenzen des "belegten" Bereichs ermitteln (Bereich, in dem sich Oberflchenpunkte befinden)
/*
  // Z: 0 -> MaxZ
  bool IstLeer = true ;
  for (Z = 0; Z < Max.z() && IstLeer; ++Z)
      for (Y = 0; Y < Max.y(); ++Y) for (X = 0; X < Max.x(); ++X)
      if (Oberflaeche[(Z*Max.y()+Y)*Max.x()+X]) IstLeer = false ;
*/
  cout << "X: von " << NeuesStartX << " bis " << NeuesEndX << endl ;
  cout << "Y: von " << NeuesStartY << " bis " << NeuesEndY << endl ;
  cout << "Z: von " << NeuesStartZ << " bis " << NeuesEndZ << endl << endl ;

  // nur weiter gehen, wenn noch mehr als 2 Spalten Platz ist
  if (NeuesStartX <= 1) NeuesStartX = 0 ;
  if (NeuesStartY <= 1) NeuesStartY = 0 ;
  if (NeuesStartZ <= 1) NeuesStartZ = 0 ;
  if (NeuesEndX >= Max.x()-2) NeuesEndX = Max.x()-1 ;
  if (NeuesEndY >= Max.y()-2) NeuesEndY = Max.y()-1 ;
  if (NeuesEndZ >= Max.z()-2) NeuesEndZ = Max.z()-1 ;

  NeuesStartX /= 2 ;
  NeuesStartY /= 2 ;
  NeuesStartZ /= 2 ;
  NeuesEndX = (Max.x()-1) - ((Max.x()-1)-NeuesEndX) /2 ;
  NeuesEndY = (Max.y()-1) - ((Max.y()-1)-NeuesEndY) /2 ;
  NeuesEndZ = (Max.z()-1) - ((Max.z()-1)-NeuesEndZ) /2 ;

  cout << "Neue Grenzen fr die nchste Iteration:" << endl ;
  cout << "X: von " << NeuesStartX << " bis " << NeuesEndX << endl ;
  cout << "Y: von " << NeuesStartY << " bis " << NeuesEndY << endl ;
  cout << "Z: von " << NeuesStartZ << " bis " << NeuesEndZ << endl ;
  cout << "X: von " << RealWert (0,NeuesStartX) << " bis " << RealWert (0,NeuesEndX) << endl ;
  cout << "Y: von " << RealWert (1,NeuesStartY) << " bis " << RealWert (1,NeuesEndY) << endl ;
  cout << "Z: von " << RealWert (2,NeuesStartZ) << " bis " << RealWert (2,NeuesEndZ) << endl << endl ;

  // neue Koordinaten ausgeben
  ostringstream NeueKoordinaten ;
  NeueKoordinaten << '"' << RealWert (0,NeuesStartX) << ' ' << RealWert (0,NeuesEndX) << ' ' 
                  << RealWert (1,NeuesStartY) << ' ' << RealWert (1,NeuesEndY) << ' '
      << RealWert (2,NeuesStartZ) << ' ' << RealWert (2,NeuesEndZ) << '"' << ends ;

  // Speicher freigeben
  delete [] Innerhalb ;
  delete [] Oberflaeche ;
  delete [] Anisotropie ;

  return (NeueKoordinaten.str()) ;
}

/**********************/
/*                    */
/* globale Funktionen */
/*                    */
/**********************/

void Hilfe () 
{ cout << endl
       << "AICD-opt_remap - optimiert die Parameter fuer AICD-remap" << endl
       << endl
       << "Dieses Programm bestimmt einen quaderfoermigen Bereich so," << endl
       << "dass bei einem waehlbaren Limit der Anisotropie der Stromdichte" << endl
       << "die Isooberflaeche gerade diesen Bereich ausfuellt." << endl
       << endl
       << "Um dies zu erreichen wird AICD-remap mehrfach aufgerufen." << endl
       << "Als Ergebnis erhaelt man somit direkt das output-file von" << endl
       << "AICD-remap." << endl
       << endl 
       << "Aufruf: AICD-opt_remap  Anzahl  limit  output-file  input-file" << endl
       << "Anzahl: Anzahl der Gitterpunkte (s. AICD-remap)" << endl 
       << "Limit: Grenzwert der Isooberflaeche (s. AICD-isosurface)" << endl 
       << "Output: s. AICD-remap" << endl 
       << "Input: s. AICD-remap" << endl 
       << endl ;
} 

/*****************/
/*               */
/* Hauptprogramm */
/*               */
/*****************/

int main (int argc, char* argv [])
{ if (argc < 5) Hilfe () ; else
  { // Anzahl der Gitterpunkte einlesen
    istringstream PunktzahlEinleser (argv[1]) ;
    int AnzahlGitterpunkte ;
    PunktzahlEinleser >> AnzahlGitterpunkte ;

  // Limit einlesen
    float Limit ;
    istringstream LimitEinleser (argv[2]) ;
    LimitEinleser >> Limit ;

    // Parameter einlesen    
    char* Inputfilename = argv[4] ;
    cout << endl << "Input-file: " << Inputfilename << endl ;

    // Parameter einlesen    
    char* Outputfilename = argv[3] ;
    cout << endl << "Output-file: " << Outputfilename << endl ;

  // erster Aufruf von NMR-kartes
  int Iteration = 1 ;
  cout << "========================" << endl
     << " " << Iteration << ". Iteration" << endl
     << "========================" << endl ;
  ostringstream Kommandozeile ;
  Kommandozeile << "AICD-remap " << AnzahlGitterpunkte << " 1 - " << Outputfilename << " " << Inputfilename << ends ;
  char** args = new char* [7] ;
  string* ZwischenStr = new string [6] ;
  istringstream ArgAusleser (Kommandozeile.str()) ;
  for (int T = 0; T < 6; ++T)
  { 
    ArgAusleser >> ZwischenStr [T] ;
    args [T] = (char*) ZwischenStr[T].c_str() ;
          cout << args [T] << ' ' ;   
  }
  cout << endl ;
  args [6] = NULL ;
  int KartesExitcode = spawnvp (_P_WAIT,"AICD-remap",args) ;
  cout << "Exitcode: " << KartesExitcode << endl ;

    // Einlesen der Daten aus dem output-file von NMR-kartes
  ifstream Outputfile (Outputfilename,ios::binary) ;
  KartNetz globalesNetz (Outputfile) ;
  globalesNetz.ParameterAusgabe (cout) ;
  string NeueGrenzen (globalesNetz.ErmittleNeueGrenzen (Limit)) ;
  string AlteGrenzen ("") ;
  Outputfile.close () ;

  // iterieren
  while (!(AlteGrenzen == NeueGrenzen))
  { // Aufruf von NMR-kartes
      cout << "========================" << endl
       << " " << ++Iteration << ". Iteration" << endl
       << "========================" << endl ;
    ZwischenStr [3] = NeueGrenzen.substr(1,NeueGrenzen.size()-2) ;
  
      for (int T = 0; T < 6; ++T) 
    { 
      args [T] = (char*) ZwischenStr[T].c_str() ;
      cout << args [T] << ' ' ;
    }
    cout << endl ;
    args [6] = NULL ;
    KartesExitcode = spawnvp (_P_WAIT,"AICD-remap",args) ;
    cout << "Exitcode: " << KartesExitcode << endl ;
    AlteGrenzen = NeueGrenzen ;

    // Einlesen der Daten aus dem output-file von NMR-kartes
    Outputfile.open (Outputfilename,ios::binary) ;
    KartNetz globalesNetz (Outputfile) ;
    globalesNetz.ParameterAusgabe (cout) ;
    NeueGrenzen = globalesNetz.ErmittleNeueGrenzen (Limit) ;
    Outputfile.close () ;
  }

  int Suchbereich = 1 ;
  while (KartesExitcode == 2)
  { // Suchbereich vergrern, bis alle Punkte interpoliert werden knnen
      cout << "========================" << endl
       << " " << ++Iteration << ". Iteration" << endl
       << "========================" << endl ;
    ostringstream Argument3 ;
    Argument3 << ++Suchbereich << ends ;
    ZwischenStr [2] = Argument3.str() ;
  
      for (int T = 0; T < 6; ++T) 
    { 
      args [T] = (char*) ZwischenStr[T].c_str() ;
      cout << args [T] << ' ' ;
          }     
    cout << endl ;
    args [6] = NULL ;
    KartesExitcode = spawnvp (_P_WAIT,"AICD-remap",args) ;
    cout << "Exitcode: " << KartesExitcode << endl ;
  }
  cout << "==========================" << endl
     << " Keine weitere Iteration" << endl
     << " notwendig. Anzahl der" << endl
     << " Iterationen: " << Iteration << endl
     << "==========================" << endl ;
  }
  return 0;
}

