#include "Tools.h"
#include "matrix.h"
#include <sstream>
#include <fstream>
#include "minmax.h"
#include <string>
#include <list>
#include <algorithm>
#include <cstdlib>
#include <cmath>

using namespace std ;

float abs (float Wert)
{ if (Wert < 0) return (-Wert) ;
     else return (Wert) ;
}

/******************/
/*                */
/* Klasse Minimax */
/*                */
/******************/

const float Gross = float (1e37) ;

class Minimax
{ private:
  float MinIso,MaxIso,MinAniso,MaxAniso ;
  float MinX,MaxX,MinY,MaxY,MinZ,MaxZ ;

  public:
  Minimax () 
  : MinIso (Gross), MaxIso (-Gross), MinAniso (Gross), MaxAniso (-Gross),
    MinX (Gross), MaxX (-Gross), MinY (Gross), MaxY (-Gross), MinZ (Gross), MaxZ (-Gross)
  {
  }
  
  float NormOrt (int Ebene,float Wert) ;
  float NormOrtUnbegrenzt (int Ebene, float Wert) ;
  float RealOrt (int Ebene,float Wert) ;
  float NormIso (float Wert) ;
  float NormAniso (float Wert) ;
  float DiffX () { return (MaxX-MinX) ; } ;
  float DiffY () { return (MaxY-MinY) ; } ;
  float DiffZ () { return (MaxZ-MinZ) ; } ;
  void SetOrtsGrenzen (float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ) ;
  void SchraenkeEin (float Delta) ;
  void CheckGrenzen (const Minimax& MM)
  { CheckGrenzen (MM.MinIso,MM.MaxIso,MM.MinAniso,MM.MaxAniso,MM.MinX,MM.MaxX,MM.MinY,MM.MaxY,MM.MinZ,MM.MaxZ) ;
  }
  void CheckGrenzen (float aMinIso,float aMaxIso,float aMinAniso,float aMaxAniso,float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ) ;
  friend ostream& operator << (ostream& OutStream, const Minimax& MM) ;
  friend istream& operator >> (istream& InStream, Minimax& MM) ;
} ;

void Minimax::SchraenkeEin (float Delta) 
{ MinX += Delta ;
  MaxX -= Delta ;
  MinY += Delta ;
  MaxY -= Delta ;
  MinZ += Delta ;
  MaxZ -= Delta ;
}

void Minimax::SetOrtsGrenzen (float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ)
{ MinX = aMinX ; MaxX = aMaxX ;
  MinY = aMinY ; MaxY = aMaxY ;
  MinZ = aMinZ ; MaxZ = aMaxZ ;
}

void Minimax::CheckGrenzen (float aMinIso,float aMaxIso,float aMinAniso,float aMaxAniso,float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ)
{ if (MinIso > aMinIso) MinIso = aMinIso ;
  if (MaxIso < aMaxIso) MaxIso = aMaxIso ;
  if (MinAniso > aMinAniso) MinAniso = aMinAniso ;
  if (MaxAniso < aMaxAniso) MaxAniso = aMaxAniso ;
  if (MinX > aMinX) MinX = aMinX ;
  if (MaxX < aMaxX) MaxX = aMaxX ;
  if (MinY > aMinY) MinY = aMinY ;
  if (MaxY < aMaxY) MaxY = aMaxY ;
  if (MinZ > aMinZ) MinZ = aMinZ ;
  if (MaxZ < aMaxZ) MaxZ = aMaxZ ;
}

float Minimax::NormIso (float Wert) 
{ if (MaxIso == MinIso) return (0) ; else 
  { float Ergebnis = (Wert-MinIso)/(MaxIso-MinIso) ;
    if (Ergebnis < 0) Ergebnis = 0 ;
    if (Ergebnis > 1) Ergebnis = 1 ;
    return (Ergebnis) ;
  }
}

float Minimax::NormAniso (float Wert) 
{ if (MaxAniso == MinAniso) return (0) ; else 
  { float Ergebnis = ((Wert)-(MinAniso))/(MaxAniso-MinAniso) ;
    if (Ergebnis < 0) Ergebnis = 0 ;
    if (Ergebnis > 1) Ergebnis = 1 ;
    return (Ergebnis) ;
    //return ((Wert-MinAniso)/(MaxAniso-MinAniso)) ;
  }
}

float Minimax::NormOrtUnbegrenzt (int Ebene,float Wert) 
{ float Ergebnis ;

  switch (Ebene)
  { case 0: { Ergebnis = (Wert-MinX)/(MaxX-MinX) ; break ; } ;
    case 1: { Ergebnis = (Wert-MinY)/(MaxY-MinY) ; break ; } ;
    case 2: { Ergebnis = (Wert-MinZ)/(MaxZ-MinZ) ; break ; } ;
  }
  return (Ergebnis) ;  
}

float Minimax::NormOrt (int Ebene,float Wert) 
{ float Ergebnis = NormOrtUnbegrenzt (Ebene,Wert) ;
  if (Ergebnis < 0) Ergebnis = 0 ;
  if (Ergebnis > 1) Ergebnis = 1 ;
  return (Ergebnis) ;  
}

float Minimax::RealOrt (int Ebene,float Wert) 
{ float Ergebnis ;
  
  if (Wert < 0) Wert = 0 ;
  if (Wert > 1) Wert = 1 ;

  switch (Ebene)
  { case 0: { Ergebnis = Wert*(MaxX-MinX)+MinX ; break ; } ;
    case 1: { Ergebnis = Wert*(MaxY-MinY)+MinY ; break ; } ;
    case 2: { Ergebnis = Wert*(MaxZ-MinZ)+MinZ ; break ; } ;
  }
  return (Ergebnis) ;
}

ostream& operator << (ostream& OutStream, const Minimax& MM) 
{ OutStream << MM.MinIso << " <= Isotropie <= " << MM.MaxIso << endl
            << MM.MinAniso << " <= Anisotropie <= " << MM.MaxAniso << endl
            << MM.MinX << " <= X <= " << MM.MaxX << endl
            << MM.MinY << " <= Y <= " << MM.MaxY << endl
            << MM.MinZ << " <= Z <= " << MM.MaxZ << endl ;
  return (OutStream) ;
}

istream& operator >> (istream& InStream, Minimax& MM) 
{ string Dummy ;

  InStream >> MM.MinIso >> Dummy >> Dummy >> Dummy >> MM.MaxIso
           >> MM.MinAniso >> Dummy >> Dummy >> Dummy >> MM.MaxAniso
           >> MM.MinX >> Dummy >> Dummy >> Dummy >> MM.MaxX
           >> MM.MinY >> Dummy >> Dummy >> Dummy >> MM.MaxY
           >> MM.MinZ >> Dummy >> Dummy >> Dummy >> MM.MaxZ ;
  return (InStream) ;
}

/*******************/
/*                 */
/* Klasse KartNetz */
/*                 */
/*******************/

class KartNetz
{ private:
  Minimax Grenzen ;
  float* Daten ;
  float* Abstaende ;
  int* Indices ;
  int AnzahlGitterpunkte ;
  Matrix<int> Max ;
  int Suchbereich ;

  public:
  KartNetz () 
  : AnzahlGitterpunkte (0), Daten (0), Abstaende (0), Indices (0), Max (Null,1,3), Suchbereich (1) {} ;
  KartNetz (int Punktzahl, Minimax aGrenzen, int aSuchbereich) ;
  ~KartNetz () ;

  void ParameterAusgabe (ostream& OutStream) const ;
  friend ostream& operator << (ostream& OutStream, const KartNetz& KN) ;

  float* At (int x,int y,int z) const ;
  int NextInt (int Koordinate, float Wert) ;
  int PrevInt (int Koordinate, float Wert) ;
  inline float RealWert (int Koordinate, int Wert) ;

  void Summieren (Matrix<float> Ort, Matrix<float> Wert, int Index) ;
  bool Division (void* GaussNetz) ;
} ;

int KartNetz::PrevInt (int Koordinate, float Wert) 
{ float TempD = Grenzen.NormOrt (Koordinate,Wert) ;
  if (TempD == 0) return (-1) ; else return (int ((*Max[Koordinate]-1) * TempD)) ;
//return (int ((*Max[Koordinate]-1) * Grenzen.NormOrt (Koordinate,Wert))) ;
}

inline float KartNetz::RealWert (int Koordinate, int Wert) 
{ return (Grenzen.RealOrt (Koordinate,float (Wert) / (*Max[Koordinate]-1))) ;
}

inline int KartNetz::NextInt (int Koordinate, float Wert) 
{ return (PrevInt (Koordinate,Wert)+1) ;

/*int TempInt = PrevInt (Koordinate,Wert)+1 ;
  if (TempInt >= *Max[Koordinate]) TempInt = *Max[Koordinate] ;
  return (TempInt) ;
*/
}

KartNetz::KartNetz (int Punktzahl, Minimax aGrenzen, int aSuchbereich) 
: AnzahlGitterpunkte (Punktzahl), Max (Null,1,3), Grenzen (aGrenzen), Suchbereich (aSuchbereich)
{ float PunkteProEinheit = float (AnzahlGitterpunkte) / (Grenzen.DiffX()*Grenzen.DiffY()*Grenzen.DiffZ()) ;
  PunkteProEinheit = exp (1.0/3.0*log (PunkteProEinheit)) ;
  *Max [0] = int (Grenzen.DiffX()*PunkteProEinheit +0.5) ;
  *Max [1] = int (Grenzen.DiffY()*PunkteProEinheit +0.5) ;
  *Max [2] = int (Grenzen.DiffZ()*PunkteProEinheit +0.5) ;

  AnzahlGitterpunkte = Max.x()*Max.y()*Max.z() ;
  Daten = new float [AnzahlGitterpunkte*9] ;
  float* EndDaten = Daten + AnzahlGitterpunkte*9 ;
  for (float* TempDaten = Daten; TempDaten < EndDaten; TempDaten++) *TempDaten = 0.0 ;

  Abstaende = new float [AnzahlGitterpunkte*4] ;
  EndDaten = Abstaende + 4*AnzahlGitterpunkte ;
  for (float* TempDaten = Abstaende; TempDaten < EndDaten; TempDaten++) *TempDaten = Gross ;
  Indices = new int [AnzahlGitterpunkte*4] ;
}

KartNetz::~KartNetz () 
{ delete [] Daten ;
  delete [] Indices ;
  delete [] Abstaende ;
}

float* KartNetz::At (int x,int y,int z) const
{ return (&Daten[((z*Max.y()+y)*Max.x()+x)*9]) ;
}

void KartNetz::ParameterAusgabe (ostream& OutStream) const
{ OutStream << AnzahlGitterpunkte << " = " << *Max[0] << " * " << *Max[1] << " * " << *Max[2] ;
}

void KartNetz::Summieren (Matrix<float> Ort, Matrix<float> Wert, int Index) 
{ // umrechnen der Ortskoordinaten in Normkoordinaten
  float XU = Grenzen.NormOrtUnbegrenzt (0,Ort.x()) ;
  float YU = Grenzen.NormOrtUnbegrenzt (1,Ort.y()) ;
  float ZU = Grenzen.NormOrtUnbegrenzt (2,Ort.z()) ;

  // testen, ob der Punkt innerhalb des KartNetzes liegt
  if ((XU >= 0) && (XU < 1) && (YU >= 0) && (YU < 1) && (ZU >= 0) && (ZU < 1))
  { // Indices berechnen
    int iX = int (Max.x()*XU) ;
    int iY = int (Max.y()*YU) ;
    int iZ = int (Max.z()*ZU) ;

    for (int zz = max(iZ-Suchbereich,0); zz < min(iZ+1+Suchbereich,Max.z()); zz++)
    { for (int yy = max(iY-Suchbereich,0); yy < min(iY+1+Suchbereich,Max.y()); yy++)
      { for (int xx = max(iX-Suchbereich,0); xx < min(iX+1+Suchbereich,Max.x()); xx++)
        { // Position im Abstandsarray ermitteln
	  float* AnfAbstaende = Abstaende + ((zz*Max.y()+yy)*Max.x()+xx)*4 ;

	  // Abstandsquadrat berechnen
	  float Abstand = (sqr((float (xx)+0.5) / float (Max.x())-XU)
			   +sqr((float (yy)+0.5) / float (Max.y())-YU)
			   +sqr((float (zz)+0.5) / float (Max.z())-ZU)) ;

	  // schauen, ob schlechteres Abstandquadrat gespeichert ist, falls ja, dann ersetzen
	  for (int Zlr = 0; Zlr < 4; ++Zlr)
	  { if (Abstand < AnfAbstaende[Zlr])
	    { int* AnfIndices = Indices + ((zz*Max.y()+yy)*Max.x()+xx)*4 ;

	      for (int Zlr2 = 3; Zlr2 > Zlr; --Zlr2)
	      { AnfIndices[Zlr2] = AnfIndices[Zlr2-1] ;
		AnfAbstaende[Zlr2] = AnfAbstaende[Zlr2-1] ;
	      }
	      AnfAbstaende[Zlr] = Abstand ;
	      AnfIndices[Zlr] = Index ;
	      break ;
	    }
	  }
	}
      }
    }
  }
}

ostream& operator << (ostream& OutStream, const KartNetz& KN) 
{ OutStream << "Gitternetzkoordinaten: " ;
  KN.ParameterAusgabe (OutStream) ;
  OutStream << endl << "Grenzen:" << endl << KN.Grenzen ;

  OutStream.write ((char *)KN.Daten,sizeof (float)*9*KN.AnzahlGitterpunkte) ;
  return (OutStream) ;
}

/******************/
/*                */
/* Klasse G94Netz */
/*                */
/******************/

class G94Netz
{ private:
  float* Daten ;
  int AnzahlGitterpunkte ;
  Matrix<float> Mittelpunkt ;
  float Radius ;

  public:
  G94Netz () : AnzahlGitterpunkte (0), Daten (0), Mittelpunkt (Null,1,3), Radius (0) {} ;
  G94Netz (istream& InStream) ;
  ~G94Netz () ;

/*  Matrix<float> Interpoliere (const Matrix<float>& Interpolationsort) ;*/
  bool Transformiere (KartNetz& KN) ;

  Matrix<float> GetMittelpunkt ()
  { return (Mittelpunkt) ; } ;
  float GetRadius ()
  { return (Radius) ; } ;
  float W (int Index, int Index2)
  { return (Daten[Index*12+3+Index2]) ; } ;
  float X (int Index)
  { return (Daten[Index*12]) ; } ;
  float Y (int Index)
  { return (Daten[Index*12+1]) ; } ;
  float Z (int Index)
  { return (Daten[Index*12+2]) ; } ;
  Matrix<float> GetWert (int Index)
  { return (Matrix<float> (&Daten[Index*12+3],3,3)) ; } ;
  float Abstand (int Index, Matrix<float> Ort)
  { Matrix<float> TempOrt (&Daten[Index*12],1,3) ;
    return ((TempOrt-Ort).Laenge ()) ;
  }
} ;

G94Netz::G94Netz (istream& InStream)
: Daten (0), Mittelpunkt (Null,1,3)
{ string Dummy ;
  InStream >> Dummy >> Dummy >> Dummy >> AnzahlGitterpunkte ;
  for (int index = 0; index < 27; ++index) InStream >> Dummy ;
  InStream >> Mittelpunkt >> Dummy >> Radius ;
  InStream.ignore (1) ;

  Daten = new float [AnzahlGitterpunkte*12] ;
  InStream.read ((char *)Daten,(signed long int) (AnzahlGitterpunkte*12*sizeof(float))) ;

/*
  cout << "fail(): " << InStream.fail () << endl ;
  cout << "gcount(): " << InStream.gcount () << endl ;
  cout << "eof(): " << InStream.eof() << endl ;
  cout << "InStream: " << InStream << endl ;
*/
/*
  for (int z = 0; z < 20; ++z) // z < AnzahlGitterpunkte
   { 
     cout << X(z) << ' ' << Y(z) << ' ' << Z(z) << ' ' 
          << W(z,0) << ' ' << W(z,1) << ' ' << W(z,2) << ' ' 
          << W(z,3)<< ' '  << W(z,4) << ' ' << W(z,5) << ' '
          << W(z,6)<< ' '  << W(z,7) << ' ' << W(z,8) << ' '
          << endl ;
   }
*/
}

G94Netz::~G94Netz () 
{ delete [] Daten ;
}

/*
Matrix<float> G94Netz::Interpoliere (const Matrix<float>& Interpolationsort)
{ // zuerst: binre Suche nach der nchstgelegenen z-Position
  int MinIndex = 0 ;
  int MaxIndex = AnzahlGitterpunkte-1 ;
  int Index = AnzahlGitterpunkte/2 -1 ;
  while (Index != MinIndex)
  { if (Interpolationsort.z() < Z(Index))
    { MaxIndex = Index ;
      Index = (MinIndex + Index) /2 ;
    } else
    if (Interpolationsort.z() > Z(Index))
    { MinIndex = Index ;
      Index = (MaxIndex + Index) /2 ;
    } else break ;
  }

  // Index zeigt nun auf die beste Annherung fr die Z-Koordinate
  // nun die vier nchsten Punkte suchen
  int Indices [4] ;
  float Abstaende [4] = {float (1e38),float (1e38),float (1e38),float (1e38)} ;

  // Index merken
  MinIndex = Index ;
  while ((Index < AnzahlGitterpunkte) && (abs(Z(Index)-Interpolationsort.z()) < Abstaende[3]))
  { float AktuellerAbstand = Abstand (Index,Interpolationsort) ;

    for (int Zlr = 0; Zlr < 4; ++Zlr)
    { if (AktuellerAbstand < Abstaende[Zlr])
      { for (int Zlr2 = 3; Zlr2 > Zlr; --Zlr2)
        { Indices[Zlr2] = Indices[Zlr2-1] ;
  	  Abstaende[Zlr2] = Abstaende[Zlr2-1] ;
	}
        Abstaende[Zlr] = AktuellerAbstand ;
        Indices[Zlr] = Index ;
        break ;
      }
    }
    ++Index ;
  }

  // nun in die andere Richtung suchen
  Index = MinIndex-1 ;
  while ((Index > 0) && (abs(Z(Index)-Interpolationsort.z()) < Abstaende[3]))
  { float AktuellerAbstand = Abstand (Index,Interpolationsort) ;

    for (int Zlr = 0; Zlr < 4; ++Zlr)
    { if (AktuellerAbstand < Abstaende[Zlr])
      { for (int Zlr2 = 3; Zlr2 > Zlr; --Zlr2)
        { Indices[Zlr2] = Indices[Zlr2-1] ;
  	  Abstaende[Zlr2] = Abstaende[Zlr2-1] ;
	}
        Abstaende[Zlr] = AktuellerAbstand ;
        Indices[Zlr] = Index ;
        break ;
      }
    }
    --Index ;
  }

  // die Werte interpolieren
  // einige Variablen initialisieren
  float x1 = X(Indices[0]) ;
  float y1 = Y(Indices[0]) ;
  float z1 = Z(Indices[0]) ;

  float x0 = Interpolationsort.x() - x1 ;
  float y0 = Interpolationsort.y() - y1 ;
  float z0 = Interpolationsort.z() - z1 ;

  float x2 = X(Indices[1]) - x1 ;
  float y2 = Y(Indices[1]) - y1 ;
  float z2 = Z(Indices[1]) - z1 ;

  float x3 = X(Indices[2]) - x1 ;
  float y3 = Y(Indices[2]) - y1 ;
  float z3 = Z(Indices[2]) - z1 ;

  float x4 = X(Indices[3]) - x1 ;
  float y4 = Y(Indices[3]) - y1 ;
  float z4 = Z(Indices[3]) - z1 ;

  float Nenner = x2*y3*z4 + x3*y4*z2 + x4*y2*z3 - x4*y3*z2 - x3*y2*z4 - x2*y4*z3 ;
  float Ergebnis [9] ;
  if (Nenner == 0)
  { // als Ergebnis die Werte vom nchsten Punkt nehmen
    for (int Zlr = 0; Zlr < 9; Zlr++) Ergebnis [Zlr] = W(Indices[0],Zlr) ;
  } else
  { float Faktor1 = x0*y2 - y0*x2 ;
    float Faktor2 = x2*z0 - x0*z2 ;
    float Faktor3 = y0*z2 - y2*z0 ;
    float Faktor4 = x0*(y3*z4 - y4*z3) - y0*(x3*z4 - x4*z3) + z0*(x3*y4 - x4*y3) ;

    // jetzt alle 9 Elemente des Tensors durchlaufen
    for (int Zlr = 0; Zlr < 9; Zlr++) 
    { float w1 = W(Indices[0],Zlr) ;
      float w2 = W(Indices[1],Zlr) - w1 ;
      float w3 = W(Indices[2],Zlr) - w1 ;
      float w4 = W(Indices[3],Zlr) - w1 ;
      Ergebnis[Zlr] = w1 + (Faktor1*(z3*w4 - z4*w3) 
			  + Faktor2*(y3*w4 - y4*w3) 
			  + Faktor3*(x3*w4 - x4*w3) 
			  + Faktor4*w2)/Nenner ;
    }
  }

  // Ergebnismatrix generieren
  Matrix<float> TempMat (Ergebnis,3,3) ;
  return (TempMat) ;
}
*/

bool G94Netz::Transformiere (KartNetz& KN) 
{ float* EndDaten = Daten + AnzahlGitterpunkte*12 ;
  int Index = 0 ;
  int ProzentZlr = AnzahlGitterpunkte/100 ;
  int ProzentZahl = 0 ;
  for (float* TempDaten = Daten; TempDaten < EndDaten; TempDaten += 12)
  { Matrix<float> Ort (TempDaten,1,3) ;
    Matrix<float> Wert (TempDaten+3,3,3) ;
    KN.Summieren (Ort,Wert,Index++) ;
    if (--ProzentZlr == 0)
    { cout << char (13) << ++ProzentZahl << " %" ;
      cout.flush () ;
      ProzentZlr = AnzahlGitterpunkte/100 ;
    }
  }
  cout << endl ;
  return (KN.Division (this)) ;
}

/*********************/
/*                   */
/* noch mal KartNetz */
/*                   */
/*********************/

bool KartNetz::Division (void* GaussNetz)
{ float* TempDaten = Daten ;
  float* EndAbstaende = Abstaende+AnzahlGitterpunkte*4 ;
  int* TempIndices = Indices ;
  int NichtInterpoliertePunkte = 0 ;

  for (float* TempAbstaende = Abstaende; TempAbstaende < EndAbstaende; TempAbstaende+=4)
  { int Index = 0 ;
    Matrix<float> Ergebnis (Null,3,3) ;
    float SummeGewichte = 0 ;

    while ((TempAbstaende[Index] != Gross) && (Index < 4))
    { float Gewicht = (TempAbstaende[Index] < 1/Gross) ? Gross : 1/TempAbstaende[Index] ;
      Ergebnis += ((G94Netz*) (GaussNetz))->GetWert (TempIndices[Index]) * Gewicht ;
      SummeGewichte += Gewicht ;

      ++Index ;
    }

    // Wenn kein Punkt gefunden worden ist:
    if (Index == 0)
    { NichtInterpoliertePunkte++ ;
    } else 
    { Ergebnis /= SummeGewichte ;
      for (int Zlr = 0; Zlr < 9; ++Zlr) TempDaten[Zlr] = Ergebnis[0][Zlr] ;
    }

    TempDaten += 9 ;
    TempIndices += 4 ;
  }

  cout << NichtInterpoliertePunkte << " von " << AnzahlGitterpunkte << " Punkten konnten nicht interpoliert werden. (" << float (NichtInterpoliertePunkte)/AnzahlGitterpunkte*100 << "%)" << endl ;
  return (NichtInterpoliertePunkte == 0) ;
}

/**********************/
/*                    */
/* globale Funktionen */
/*                    */
/**********************/

void Hilfe () 
{ cout << endl 
       << "AICD-remap - Transformation der Stromdichtetensoren auf ein" << endl
       << "             kartesisches Koordinatennetz" << endl
       << endl
       << "Dieses Programm transformiert die Daten, die AICD-extract als" << endl
       << "Output produziert, auf ein kartesisches Koordinatennetz." << endl
       << endl
       << "Das Koordinatennetz wird durch die gewuenschte Anzahl der enthaltenen" << endl
       << "Gitterpunkte und die Grenzen der Koordinaten bestimmt. Daraus" << endl
       << "wird die Anzahl der Gitterpunkte in jeder Raumrichtung bestmoeglich" << endl
       << "ermittelt." << endl
       << endl
       << "Da eventuell nicht alle Tensoren auf dem Gitternetz interpoliert werden" << endl
       << "koennen, wird der Anteil der ermittelten Tensoren ausgegeben. Ist dieser" << endl
       << "Anteil 0, so beendet sich das Programm mit einem exit-code von 0, andern-" << endl
       << "falls ist der exit-code 2." << endl
       << endl
       << "Das output-file hat ein binaeres Format. Informationen ueber das Gitter-" << endl
       << "netz koennen allerdings mit Hilfe des Kommandos" << endl
       << endl
       << "head -7 output-file" << endl
       << endl
       << "ausgegeben werden." << endl
       << endl
       << "Aufruf: AICD-remap  Anzahl  Suchbereich  Grenzen  output-file  input-file" << endl
       << "Anzahl: gewuenschte Anzahl der Gitterpunkte (z.B. 40000)" << endl
       << "Suchbereich: natuerliche Zahl, die angibt, bis zu welchem Abstand (in" << endl
       << "             Gitterpunkten gezaehlt) Tensoren zur Interpolation heran-" << endl
       << "             gezogen werden sollen. Im allgemeinen ist ein Wert von" << endl 
       << "             1 oder 2 eine gute Wahl." << endl 
       << "Grenzen: '-' bedeutet Originalgrenzen" << endl
       << "         '1 2 3 4 5 6' bedeutet:" << endl
       << "         1 <= X <= 2 und 3 <= Y <= 4 und 5 <= Z <= 6" << endl 
       << "         Achtung: die einzelnen Werte muessen mit Leerzeichen getrennt und" << endl
       << "         in Haekchen (\') eingeschlossen werden, da die Grenzen immer als" << endl
       << "         _ein_ Kommandozeilenargument gelesen werden." << endl
       << "output-file: Dateiname des output-files von AICD-remap" << endl
       << "input-file: Dateiname des output-files von AICD-extract (endet im allgemeinen" << endl
       << "            auf \".b\")" << endl
       << "Exitcodes: 0 = alle Punkte konnten interpoliert werden" << endl 
       << "           1 = falsche Anzahl von Parametern" << endl 
       << "           2 = es konnten nicht alle Punkte interpoliert werden" << endl 
       << endl ;
} 

/*****************/
/*               */
/* Hauptprogramm */
/*               */
/*****************/

int main (int argc, char* argv [])
{ char* Inputfilename ;
//  int Ebene ;
//  float von, bis ;

//          cout << argc << " Argumente" << endl ;
//	  for (int T = 0; T < argc; T++) cout << T << ". Argument = " << argv[T] << endl ;

  if (argc != 6) 
  { Hilfe () ;
    return (1) ;
  }
  else
  { // von allen Inputfiles die globalen Grenzen ermitteln

    // globale Grenzen initialisieren
    Minimax globaleGrenzen ;
    
    Inputfilename = argv[5] ;
#ifdef LINUX    
    ifstream Inputfile (Inputfilename) ;
#else
#ifdef IRIX
    ifstream Inputfile (Inputfilename) ;
#else
    ifstream Inputfile (Inputfilename,ios::binary) ;
#endif    
#endif
    cout << endl << "Input-file: " << Inputfilename << endl ;

    // Anzahl der Tensoren einlesen
    string Dummy ;
    int TensorenZahl ;
    Inputfile >> Dummy >> Dummy >> Dummy >> TensorenZahl ;
    cout << "Anzahl der Stromdichtetensoren: " << TensorenZahl << endl ;

    // Grenzen einlesen
    Minimax Grenzen ;
    Inputfile >> Dummy >> Grenzen ;
    // alle Werte um 10 verschieben, da sonst der Bereich unntig gro wird
	Grenzen.SchraenkeEin (float (10.0)) ;
    globaleGrenzen.CheckGrenzen (Grenzen) ;
    cout << "Grenzen:" << endl << Grenzen << endl ;      

    // ggf. globale Grenzen an Benutzerangabe anpassen
    if (string (argv[3]) != "-")
    { istringstream GrenzenEinleser (argv[3]) ;
      float GMinX,GMaxX,GMinY,GMaxY,GMinZ,GMaxZ ;
      GrenzenEinleser >> GMinX >> GMaxX >> GMinY >> GMaxY >> GMinZ >> GMaxZ ;
      globaleGrenzen.SetOrtsGrenzen (GMinX,GMaxX,GMinY,GMaxY,GMinZ,GMaxZ) ;
    }

    // globale Grenzen ausgeben
    cout << endl << "Grenzen des kartesischen Netzes:" << endl << globaleGrenzen << endl ;

    // kartesisches Gitternetz initialisieren (globales Netz)
    istringstream PunktzahlEinleser (argv[1]) ;
    int AnzahlGitterpunkte ;
    PunktzahlEinleser >> AnzahlGitterpunkte ;

    istringstream SuchbereichsEinleser (argv[2]) ;
    int Suchbereich ;
    SuchbereichsEinleser >> Suchbereich ;

    KartNetz globalesNetz (AnzahlGitterpunkte,globaleGrenzen,Suchbereich) ;
    cout << "Globales Netz: " ;
    globalesNetz.ParameterAusgabe (cout) ;
    cout  << endl ;

    // Tensoren remappen
    cout << endl << "Input-file: " << Inputfilename << endl ;
    Inputfile.seekg (0) ;

    // lokale Netze initialisieren
    G94Netz GaussNetz (Inputfile) ;

    // Transformation durchfhren
	int KartesExitcode ;
    if (GaussNetz.Transformiere (globalesNetz)) KartesExitcode = 0 ; else KartesExitcode = 2 ;

    // globales Netz in eine Datei schreiben
#ifdef LINUX    
    ofstream Outputfile (argv[4]) ;
#else
#ifdef IRIX
    ofstream Outputfile (argv[4]) ;
#else
    ofstream Outputfile (argv[4],ios::binary) ;
#endif    
#endif
    Outputfile << globalesNetz ;

	return (KartesExitcode) ;
  }
}

