#include "matrix.h"
#include <fstream>
#include <sstream>
#include <iostream>
#include <string>
#include <algorithm>
#include <cmath>

const double PI = double (3.14159265358979323844) ;

using namespace std ;

/**********************/
/*                    */
/* globale Funktionen */
/*                    */
/**********************/

void Hilfe () 
{ cout << endl 
       << "AICD-rotate_mol - Vorbereitung eines POV-Ray-input-files fuer eine Animation." << endl
       << endl
       << "Mit Hilfe dieses Programms wird eine Reihe von \"translate\"- und" << endl
       << "\"rotate\"-Befehlen erzeugt und in das POV-Ray-input-file eingefuegt," << endl
       << "so dass das Molekuel an eine definierte Raumposition in einer definierten" << endl
       << "Betrachtungsrichtung transformiert wird. An jede Zeile in der Inputdatei," << endl
       << "die folgendermassen lautet:" << endl
       << endl
       << "// automatische_Geometrieanpassung_einfuegen" << endl 
       << endl
       << "werden die erzeugten Befehle angehaengt." << endl
       << endl
       << "Die Transformation erfolgt anhand der Positionen dreier Atome. Nach der" << endl
       << "Transformation befindet sich das erste Atom im Ursprung, das zweite befindet" << endl
       << "sich auf dem positiven Teil der X-Achse, waehrend das dritte Atom sich in" << endl
       << "der X-Y-Ebene wiederfindet. Die Y-Koordinate des dritten Atoms ist nach der" << endl
       << "Abbildung immer positiv. Die urspruenglichen Positionen der Atome ent-" << endl
       << "nimmt das Programm dem angegebenen pdb-file." << endl
       << endl
       << "Aufruf: AICD-rotate_mol  pdb-file  Atom-Nr.1  Atom-Nr.2  Atom-Nr.3" << endl
       << "StdIn: POV-Ray-input-file wird erwartet" << endl
       << "StdOut: POV-Ray-output-file wird ausgegeben" << endl << endl
       << "Die Atome mit dem Index Atom-Nr.1, Atom-Nr.2 und Atom-Nr.3 werden zur" << endl
       << "Bestimmung der Transformation herangezogen." << endl
       << endl ;
}

/*****************/
/*               */
/* Hauptprogramm */
/*               */
/*****************/

int main (int argc, char* argv [])
{ if (argc != 5) Hilfe () ; 
  else 
  { char* PDBFilename = argv[1] ;

  int Atom [3] ;
  Matrix<double> AtomPos [3] ;
  for (int I = 0; I < 3; ++I) AtomPos[I] = Matrix<double> (Null,1,3) ;
  istringstream AtomNrEinleser ((string(argv[2])+' '+string(argv[3])+' '+string(argv[4])).c_str()) ;
  AtomNrEinleser >> Atom[0] >> Atom[1] >> Atom[2] ;

  // Atom-Nummern sortieren
//  sort (&Atom[0],&Atom[3]) ;
//  cerr << "Atomnummern (sortiert): " << Atom[0] << "  " << Atom[1] << "  " << Atom[2] << endl ;

  // Einlesen der Atom-Daten aus dem pdb-input-file
  ifstream PDBFile (PDBFilename) ;

  // Einlesen des PDB-Files
  string Dummy ;
//  double Bohr = double (0.529177249) ;

  int Zlr = 0 ;
  while (PDBFile)
  { 
    PDBFile >> Dummy ;
    while ((Dummy != "ATOM") && PDBFile) PDBFile >> Dummy ;
    if (PDBFile) 
    { 
      ++Zlr ;
      for (int I = 0; I < 3; ++I) if (Zlr == Atom[I])
      { 
        PDBFile >> Dummy >> Dummy >> Dummy >> Dummy >> AtomPos[I] ;
      }
    } 
  }



      Matrix<double> TestPos1 = AtomPos[1] ;
      Matrix<double> TestPos2 = AtomPos[2] ;
      
  // Geometrieanpassung vorbereiten
      // 1. Verschiebung um -C1 (AtomPos[0])
	  // 1a. Verschiebung ausgeben
      ostringstream GeoAnpass ;
	  GeoAnpass << "translate < " << -AtomPos[0].x() << ',' << -AtomPos[0].y() << ',' << -AtomPos[0].z() << " >" << endl ;
	  // 1b. Verschiebung durchfhren
	  AtomPos[1] -= AtomPos[0] ;
	  AtomPos[2] -= AtomPos[0] ;

	  // 2. Um die X-Achse drehen, bis die Z-Komponente von C2 verschwindet
	  // 2a. Winkel berechnen
	  double WinkelRad = (AtomPos[1].y() == 0) ? PI/2 : -atan(AtomPos[1].z()/AtomPos[1].y()) ;

      // 3. Wenn die Y-Komponente von C2 jetzt negativ ist, dann um 180 mehr drehen
	  // 3a. Winkel berechnen
	  if (AtomPos[1].y() * cos (WinkelRad) - AtomPos[1].z() * sin (WinkelRad) < 0) WinkelRad += PI ;
	  // 3b. Drehung ausgeben
	  GeoAnpass << "rotate x*" << WinkelRad*180/PI << endl ;
    // 3c. Drehung durchfhren
      Matrix<double> Abbildung = RotationsMatrix (0,WinkelRad) ;
      
      double Y2 = AtomPos[1].y() * cos (WinkelRad) - AtomPos[1].z() * sin (WinkelRad) ;
      double Y3 = AtomPos[2].y() * cos (WinkelRad) - AtomPos[2].z() * sin (WinkelRad) ;
      double Z2 = AtomPos[1].y() * sin (WinkelRad) + AtomPos[1].z() * cos (WinkelRad) ;
      double Z3 = AtomPos[2].y() * sin (WinkelRad) + AtomPos[2].z() * cos (WinkelRad) ;

      *AtomPos[1][1] = Y2 ;
      *AtomPos[1][2] = Z2 ;
      *AtomPos[2][1] = Y3 ;
      *AtomPos[2][2] = Z3 ;
      
//      cerr << AtomPos[1] ;
//      cerr << Abbildung*TestPos1 << endl ;
//      cerr << AtomPos[2] ;
//      cerr << Abbildung*TestPos2 << endl ;

//	  cerr << "AtomPos[1] " << AtomPos[1] << endl ;
//	  cerr << "AtomPos[2] " << AtomPos[2] << endl ;

	  // 4. Um die Z-Achse drehen, bis die Y-Komponente von C2 verschwindet
	  // 4a. Winkel berechnen
	  WinkelRad = (AtomPos[1].x() == 0) ? PI/2 : -atan(AtomPos[1].y()/AtomPos[1].x()) ;

	  // 5. Wenn die X-Komponente von C2 jetzt negativ ist, dann um 180 mehr drehen
	  // 5a. Winkel berechnen
	  if (AtomPos[1].x() * cos (WinkelRad) - AtomPos[1].y() * sin (WinkelRad) < 0) WinkelRad += PI ;
	  // 5b. Drehung ausgeben
	  GeoAnpass << "rotate z*" << WinkelRad*180/PI << endl ;
	  // 5c. Drehung durchfhren
	  Abbildung = RotationsMatrix (2,WinkelRad) * Abbildung ;
	  
	  double X2 = AtomPos[1].x() * cos (WinkelRad) - AtomPos[1].y() * sin (WinkelRad) ;
	  double X3 = AtomPos[2].x() * cos (WinkelRad) - AtomPos[2].y() * sin (WinkelRad) ;
             Y2 = AtomPos[1].x() * sin (WinkelRad) + AtomPos[1].y() * cos (WinkelRad) ;
             Y3 = AtomPos[2].x() * sin (WinkelRad) + AtomPos[2].y() * cos (WinkelRad) ;
      *AtomPos[1][0] = X2 ;
      *AtomPos[2][0] = X3 ;
      *AtomPos[1][1] = Y2 ;
      *AtomPos[2][1] = Y3 ;

//      cerr << AtomPos[1] ;
//      cerr << Abbildung*TestPos1 << endl ;
//      cerr << AtomPos[2] ;
//      cerr << Abbildung*TestPos2 << endl ;


//	  cerr << "AtomPos[1] " << AtomPos[1] << endl ;
//	  cerr << "AtomPos[2] " << AtomPos[2] << endl ;

	  // 6. Um die X-Achse drehen, bis die Z-Komponente von C3 verschwindet
	  // 6a. Winkel berechnen	  
	  WinkelRad = (AtomPos[2].y() == 0) ? PI/2 : -atan(AtomPos[2].z()/AtomPos[2].y()) ;

	  // 7. Wenn die Y-Komponente von C3 jetzt negativ ist, dann um 180 mehr drehen
	  // 7a. Winkel berechnen
	  if (AtomPos[2].y() * cos (WinkelRad) - AtomPos[2].z() * sin (WinkelRad) < 0) WinkelRad += PI ;
	  // 7b. Drehung ausgeben
	  GeoAnpass << "rotate x*" << WinkelRad*180/PI << endl ;
	  // 7c. Drehung durchfhren
	  Abbildung = RotationsMatrix (0,WinkelRad) * Abbildung ;
	  
      Y2 = AtomPos[1].y() * cos (WinkelRad) - AtomPos[1].z() * sin (WinkelRad) ;
      Y3 = AtomPos[2].y() * cos (WinkelRad) - AtomPos[2].z() * sin (WinkelRad) ;
      Z2 = AtomPos[1].y() * sin (WinkelRad) + AtomPos[1].z() * cos (WinkelRad) ;
      Z3 = AtomPos[2].y() * sin (WinkelRad) + AtomPos[2].z() * cos (WinkelRad) ;
      *AtomPos[1][1] = Y2 ;
      *AtomPos[2][1] = Y3 ;
      *AtomPos[1][2] = Z2 ;
      *AtomPos[2][2] = Z3 ;
	  
      AtomPos[1] = TestPos1 ;
      AtomPos[2] = TestPos2 ;
	  GeoAnpass << "// Atomnummern: " << Atom[0] << "  " << Atom[1] << "  " << Atom[2] << endl 
                << "/* Abbildungsmatrix:" << endl
				<< Abbildung ;
/*	  for (int Index = 0; Index < 3; ++Index)
      GeoAnpass << "   Ortsvektor von Atom " << Atom[Index] << " (vor der Abbildung):" << endl
				<< AtomPos[Index] << endl
                << "   Ortsvektor von Atom " << Atom[Index] << " (nach der Abbildung):" << endl
				<< Abbildung*(AtomPos[Index]-AtomPos[0]) << endl ;
*/				
	  GeoAnpass << "*/" << endl ;
	  GeoAnpass << ends ;

//      cerr << AtomPos[1] ;
//      cerr << Abbildung*TestPos1 << endl ;
//      cerr << AtomPos[2] ;
//      cerr << Abbildung*TestPos2 << endl ;
      
      // Jetzt alle Atome einlesen und die Extrempositionen nach der Transformation merken
      
      

//	  cerr << "AtomPos[1] " << AtomPos[1] << endl ;
//	  cerr << "AtomPos[2] " << AtomPos[2] << endl ;



  // zeilenweise von StdIn lesen und zeilenweise ausgeben.
  // wenn die Zeile den Text "automatische_Geometrieanpassung_einfuegen"
  // enthlt, dann an diese Zeile die Geometrieanpassung anhngen
  char Zeile [500] ;
  string TestString ;
  bool TestErgebnis ;
  
  // F.K.: make compatible to 3.5
  cout << "#version 3.1 ;" << endl;
  
  while (cin)
  { 
    cin.getline (Zeile,499) ;
    cout << Zeile << endl ;

    // jetzt auf das Schlsselwort testen
    istringstream SchlWortTester (Zeile) ;
	TestErgebnis = false ;
    SchlWortTester >> TestString ;
    while (SchlWortTester)
    { 
      if (TestErgebnis = (TestString == "automatische_Geometrieanpassung_einfuegen"))
	  { 
	    break ; 
	  } ;
      SchlWortTester >> TestString ;
    }

    // ggf. Geometrieanpassung ausgeben
    if (TestErgebnis) 
	{
	  cout << GeoAnpass.str() ;
	}
  }
  }
  return 0;
}
