//#define LINUX
//#define IRIX
//#include "trace.h"
#include "Tools.h"
#include "matrix.h"
#include <sstream>
#include <fstream>
#include <iomanip>
#include "minmax.h"
#include <string>
#include <list>
#include <set>
#include <vector>
#include <algorithm>
#include <cstdlib>
#include <cmath>

using namespace std ;

/******************/
/*                */
/* Klasse Minimax */
/*                */
/******************/

const float Gross = float (1e38) ;

class Minimax
{ private:
  float MinIso,MaxIso,MinAniso,MaxAniso ;
  float MinX,MaxX,MinY,MaxY,MinZ,MaxZ ;

  public:
  Minimax () 
  : MinIso (Gross), MaxIso (-Gross), MinAniso (Gross), MaxAniso (-Gross),
    MinX (Gross), MaxX (-Gross), MinY (Gross), MaxY (-Gross), MinZ (Gross), MaxZ (-Gross)
  {
  }
  
  float NormOrt (int Ebene,float Wert) ;
  float RealOrt (int Ebene,float Wert) ;
  float NormIso (float Wert) ;
  float NormAniso (float Wert) ;
  float DiffX () { return (MaxX-MinX) ; } ;
  float DiffY () { return (MaxY-MinY) ; } ;
  float DiffZ () { return (MaxZ-MinZ) ; } ;
  void CheckGrenzen (const Minimax& MM)
  { CheckGrenzen (MM.MinIso,MM.MaxIso,MM.MinAniso,MM.MaxAniso,MM.MinX,MM.MaxX,MM.MinY,MM.MaxY,MM.MinZ,MM.MaxZ) ;
  }
  void CheckGrenzen (float aMinIso,float aMaxIso,float aMinAniso,float aMaxAniso,float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ) ;
  friend ostream& operator << (ostream& OutStream, const Minimax& MM) ;
  friend istream& operator >> (istream& InStream, Minimax& MM) ;
} ;

void Minimax::CheckGrenzen (float aMinIso,float aMaxIso,float aMinAniso,float aMaxAniso,float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ)
{ if (MinIso > aMinIso) MinIso = aMinIso ;
  if (MaxIso < aMaxIso) MaxIso = aMaxIso ;
  if (MinAniso > aMinAniso) MinAniso = aMinAniso ;
  if (MaxAniso < aMaxAniso) MaxAniso = aMaxAniso ;
  if (MinX > aMinX) MinX = aMinX ;
  if (MaxX < aMaxX) MaxX = aMaxX ;
  if (MinY > aMinY) MinY = aMinY ;
  if (MaxY < aMaxY) MaxY = aMaxY ;
  if (MinZ > aMinZ) MinZ = aMinZ ;
  if (MaxZ < aMaxZ) MaxZ = aMaxZ ;
}

float Minimax::NormIso (float Wert) 
{ if (MaxIso == MinIso) return (0) ; else 
  { float Ergebnis = (Wert-MinIso)/(MaxIso-MinIso) ;
    if (Ergebnis < 0) Ergebnis = 0 ;
    if (Ergebnis > 1) Ergebnis = 1 ;
    return (Ergebnis) ;
  }
}

float Minimax::NormAniso (float Wert) 
{ if (MaxAniso == MinAniso) return (0) ; else 
  { float Ergebnis = ((Wert)-(MinAniso))/(MaxAniso-MinAniso) ;
    if (Ergebnis < 0) Ergebnis = 0 ;
    if (Ergebnis > 1) Ergebnis = 1 ;
    return (Ergebnis) ;
    //return ((Wert-MinAniso)/(MaxAniso-MinAniso)) ;
  }
}

float Minimax::NormOrt (int Ebene,float Wert) 
{ float Ergebnis ;

  switch (Ebene)
  { case 0: { Ergebnis = (Wert-MinX)/(MaxX-MinX) ; break ; } ;
    case 1: { Ergebnis = (Wert-MinY)/(MaxY-MinY) ; break ; } ;
    case 2: { Ergebnis = (Wert-MinZ)/(MaxZ-MinZ) ; break ; } ;
  }
  if (Ergebnis < 0) Ergebnis = 0 ;
  if (Ergebnis > 1) Ergebnis = 1 ;
  return (Ergebnis) ;  
}

float Minimax::RealOrt (int Ebene,float Wert) 
{ float Ergebnis ;
  
  if (Wert < 0) Wert = 0 ;
  if (Wert > 1) Wert = 1 ;

  switch (Ebene)
  { case 0: { Ergebnis = Wert*(MaxX-MinX)+MinX ; break ; } ;
    case 1: { Ergebnis = Wert*(MaxY-MinY)+MinY ; break ; } ;
    case 2: { Ergebnis = Wert*(MaxZ-MinZ)+MinZ ; break ; } ;
  }
  return (Ergebnis) ;
}

ostream& operator << (ostream& OutStream, const Minimax& MM) 
{ OutStream << MM.MinIso << " <= Isotropie <= " << MM.MaxIso << endl
            << MM.MinAniso << " <= Anisotropie <= " << MM.MaxAniso << endl
            << MM.MinX << " <= X <= " << MM.MaxX << endl
            << MM.MinY << " <= Y <= " << MM.MaxY << endl
            << MM.MinZ << " <= Z <= " << MM.MaxZ << endl ;
  return (OutStream) ;
}

istream& operator >> (istream& InStream, Minimax& MM) 
{ string Dummy ;

  InStream >> MM.MinIso >> Dummy >> Dummy >> Dummy >> MM.MaxIso
           >> MM.MinAniso >> Dummy >> Dummy >> Dummy >> MM.MaxAniso
           >> MM.MinX >> Dummy >> Dummy >> Dummy >> MM.MaxX
           >> MM.MinY >> Dummy >> Dummy >> Dummy >> MM.MaxY
           >> MM.MinZ >> Dummy >> Dummy >> Dummy >> MM.MaxZ ;
  return (InStream) ;
}

/******************/
/*                */
/* Klasse Dreieck */
/*                */
/******************/

class Dreieck
{ private:
  Matrix<float> A,B,C ;
  float *AN,*BN,*CN ;
  Matrix<float> DreiecksNormale ;

  public:
  Dreieck () : A(Null,1,3), B(Null,1,3), C(Null,1,3), AN(0), BN(0), CN(0), DreiecksNormale(Null,1,3) {} ;
  Dreieck (Matrix<float> a, Matrix<float> b, Matrix<float> c, 
           float* an, float* bn, float* cn) : A(a), B(b), C(c), AN(an), BN(bn), CN(cn) 
  {
    DreiecksNormale = (A-b).KreuzProdukt (A-C) ;
    DreiecksNormale /= DreiecksNormale.Laenge () ;
    if (( ((DreiecksNormale.SkalarProdukt (Matrix<float> (AN,1,3))) >= 0 ? 1 : 0) +
             ((DreiecksNormale.SkalarProdukt (Matrix<float> (BN,1,3))) >= 0 ? 1 : 0) +
             ((DreiecksNormale.SkalarProdukt (Matrix<float> (CN,1,3))) >= 0 ? 1 : 0) ) < 2 )
    {
      DreiecksNormale = -DreiecksNormale ;
    }
  }
  Matrix<float> GetNormale () const { return (DreiecksNormale) ; } ;

//  bool operator == (const Dreieck& D) { return (A == D.A && B == D.B && C == D.C) ; } ;
//  bool operator < (const Dreieck& D) { return (A.x() < D.A.x()) ; } ;

  friend ostream& operator << (ostream& OutStream, const Dreieck& D) ;
} ;

ostream& operator << (ostream& OutStream, const Dreieck& D) 
{ // ein Bohr in Angstrm:
  float Bohr = float (0.529177249) ;
  
  // Dreieck nur ausgeben, wenn es nicht degeneriert ist!
  ostringstream TestA, TestB, TestC, TestAN, TestBN, TestCN ;
  TestA << '<' << setprecision (5) << D.A.x()*Bohr 
        << ',' << setprecision (5) << D.A.y()*Bohr
        << ',' << setprecision (5) << D.A.z()*Bohr << '>';
  TestB << '<' << setprecision (5) << D.B.x()*Bohr 
        << ',' << setprecision (5) << D.B.y()*Bohr 
        << ',' << setprecision (5) << D.B.z()*Bohr << '>' ;
  TestC << '<' << setprecision (5) << D.C.x()*Bohr 
        << ',' << setprecision (5) << D.C.y()*Bohr 
        << ',' << setprecision (5) << D.C.z()*Bohr << '>' ;
        
  Matrix<float> NormA (D.AN,1,3) ;
  Matrix<float> NormB (D.BN,1,3) ;
  Matrix<float> NormC (D.CN,1,3) ;
        
  TestAN << '<' << setprecision (5) << NormA.x()
         << ',' << setprecision (5) << NormA.y()
         << ',' << setprecision (5) << NormA.z() << '>' ;
  TestBN << '<' << setprecision (5) << NormB.x() 
         << ',' << setprecision (5) << NormB.y() 
         << ',' << setprecision (5) << NormB.z() << '>' ;
  TestCN << '<' << setprecision (5) << NormC.x() 
         << ',' << setprecision (5) << NormC.y() 
         << ',' << setprecision (5) << NormC.z() << '>' ;
        
  string SA (TestA.str ()) ;
  string SB (TestB.str ()) ;
  string SC (TestC.str ()) ;

  string SAN (TestAN.str ()) ;
  string SBN (TestBN.str ()) ;
  string SCN (TestCN.str ()) ;

  if ((!(SA == SB)) && (!(SB == SC)) && (!(SA == SC)))
  {
//    OutStream << "  triangle { " << TestA.str() << ',' << TestB.str() << ',' << TestC.str() << " }" << endl ;
    OutStream << "  smooth_triangle { " << SA << ',' << SAN << ',' << SB << ',' << SBN << ',' << SC << ',' << SCN << " }" << endl ;
  } else
  { 
    cout << "Entartetes Dreieck: " << TestA.str() << ',' << TestB.str() << ',' << TestC.str() << endl ;
  }
  
//  OutStream << "  triangle { <" << setprecision (5) << D.A.x()*Bohr << ',' << setprecision (5) << D.A.y()*Bohr << ',' << setprecision (5) << D.A.z()*Bohr 
//            << ">,<" << setprecision (5) << D.B.x()*Bohr << ',' << setprecision (5) << D.B.y()*Bohr << ',' << setprecision (5) << D.B.z()*Bohr
//            << ">,<" << setprecision (5) << D.C.x()*Bohr << ',' << setprecision (5) << D.C.y()*Bohr << ',' << setprecision (5) << D.C.z()*Bohr
//            << "> }" << endl ;
  return (OutStream) ;
}

/****************/
/*              */
/* Klasse Pfeil */
/*              */
/****************/

enum PfeilEnumTyp { A_konstant, B_konstant } ;

class Pfeil
{ private:
  Matrix<float> A,B ;
  PfeilEnumTyp FesterPunkt ;

  public:
  Pfeil () : A(Null,1,3), B(Null,1,3), FesterPunkt (A_konstant) {} ;
  Pfeil (Matrix<float> a, Matrix<float> b, PfeilEnumTyp aFesterPunkt) 
  : A(a), B(b), FesterPunkt (aFesterPunkt) {}

  Pfeil& operator /= (const float Divisor) 
  { if (FesterPunkt == A_konstant) 
    { B = A + (B-A)/Divisor ;
    } else 
    { A = B - (B-A)/Divisor ;
    }
    return (*this) ;
  } ;
  float Laenge () { return ((B-A).Laenge ()) ; } ;

  bool operator == (const Pfeil& P) { return (A == P.A && B == P.B && FesterPunkt == P.FesterPunkt) ; } ;
  bool operator < (const Pfeil& P) { return (A.x() < P.A.x()) ; } ;

  friend ostream& operator << (ostream& OutStream, const Pfeil& P) ;
} ;

ostream& operator << (ostream& OutStream, const Pfeil& P) 
{ // ein Bohr in Angstrm:
  const float Bohr = float (0.529177249) ;
  const float Pfeilspitzenlaenge = float (0.166666) ;

  Matrix<float> Mittelpunkt (Null,1,3) ;
  Matrix<float> Differenz (Null,1,3) ;
  Differenz = P.B-P.A ;
  float Schaftlaenge = Differenz.Laenge() - Pfeilspitzenlaenge ;
  if (Schaftlaenge < 0) Schaftlaenge = Differenz.Laenge()*0.01 ;
  Mittelpunkt = P.A + Differenz/Differenz.Laenge()*Schaftlaenge ;

    if ((Mittelpunkt-P.A).SkalarProdukt(P.B-Mittelpunkt) < 0) 
       OutStream << "Hier stimmt was nicht: " << (Mittelpunkt-P.A).SkalarProdukt(P.B-Mittelpunkt) << endl 
                 << Mittelpunkt-P.A << "L:  " << (Mittelpunkt-P.A).Laenge () << endl 
                 << P.B-Mittelpunkt << "L:  " << (P.B-Mittelpunkt).Laenge () << endl 
                 << "Differenz: " << Differenz << "L:  " << Differenz.Laenge () << endl 
                 << "P.A: " << P.A << "L:  " << P.A.Laenge () << endl 
                 << "P.B: " << P.B << "L:  " << P.B.Laenge () << endl 
                 << "Schaftlnge: " << Schaftlaenge << "  Pfeilspitzenlaenge: " << Pfeilspitzenlaenge << endl ;

  if ((Mittelpunkt != P.A) && (Mittelpunkt != P.B)) 
  { ostringstream TestA, TestB, TestC ;
    TestA << '<' << setprecision (5) << P.A.x()*Bohr 
          << ',' << setprecision (5) << P.A.y()*Bohr 
          << ',' << setprecision (5) << P.A.z()*Bohr << ">" ;
    TestB << '<' << setprecision (5) << Mittelpunkt.x()*Bohr 
          << ',' << setprecision (5) << Mittelpunkt.y()*Bohr 
          << ',' << setprecision (5) << Mittelpunkt.z()*Bohr << ">" ;
    TestC << '<' << setprecision (5) << P.B.x()*Bohr 
          << ',' << setprecision (5) << P.B.y()*Bohr 
          << ',' << setprecision (5) << P.B.z()*Bohr << ">" ;

    string SA (TestA.str ()) ;
    string SB (TestB.str ()) ;
    string SC (TestC.str ()) ;

    if ((!(SA == SB)) && (!(SB == SC)) && ((P.A-Mittelpunkt).Laenge() >= 2e-4))
    {
      OutStream << "  cylinder { "
                << SA.c_str() << ","
                << SB.c_str()
                << ",0.0104 pigment { Green }}"
                << " cone { " << TestB.str().c_str() 
                << ",0.0208," << TestC.str().c_str()
                << ",0 pigment { Red }}"
                << "  // Lnge: " << (P.A-Mittelpunkt).Laenge () // Debug!!!
                << endl ;
    } else
    { if (SA == SB)
      {
        cout << "Entarteter Zylinder: " << SA.c_str() << ','
             << SB.c_str() << endl ;
      }
      if (SB == SC)
      {
        cout << "Entarteter Kegel: " << SB.c_str()
             << ',' << SC.c_str() << endl ;
      }
    }
  }
  return (OutStream) ;
}

/*******************/
/*                 */
/* Klasse KartNetz */
/*                 */
/*******************/

typedef vector<Dreieck> TDreiecke ;
typedef list<Pfeil> TPfeile ;
enum IsooberflaechenArt { ICD,Aniso,AntiSymmAniso,FullAniso,Iso } ;

class KartNetz
{ private:
  Minimax Grenzen ;
  float* Daten ;
  int AnzahlGitterpunkte ;
  Matrix<int> Max ;

  public:
  KartNetz () : AnzahlGitterpunkte (0), Daten (0), Max (Null,1,3) {} ;
  KartNetz (istream& InStream) ;
  ~KartNetz () ;

  void ParameterAusgabe (ostream& OutStream) const ;
  void SchreibePov (ostream& OutStream, float AnisotropieLimit, Matrix<float> Magnetfeld, IsooberflaechenArt IOArt) ;

  float* At (int x,int y,int z) ;
  inline float RealWert (int Koordinate, int Wert) ;
} ;

KartNetz::KartNetz (istream& InStream)
: AnzahlGitterpunkte (0), Max (Null,1,3)
{ string Dummy ;

  InStream >> Dummy >> AnzahlGitterpunkte >> Dummy >> *Max[0] >> Dummy >> *Max[1] >> Dummy >> *Max[2] 
           >> Dummy >> Grenzen ;
  InStream.ignore (1) ;

//  cout << "Streampos: " << InStream.tellg () << endl ;

  assert (AnzahlGitterpunkte == Max.x()*Max.y()*Max.z()) ;
  Daten = new float [AnzahlGitterpunkte*9] ;
  InStream.read ((char *) Daten,sizeof (float)*9*AnzahlGitterpunkte) ;

/*
     for (int Index = 0; Index < 10; Index++)
     { for (int Index2 = 0; Index2 < 9; Index2++)
           cout << Daten[Index*9+Index2] << "  " ;
       cout << endl ;
     }
*/
}

KartNetz::~KartNetz ()
{ delete [] Daten ;
}

inline float KartNetz::RealWert (int Koordinate, int Wert)
{ return (Grenzen.RealOrt (Koordinate,float (Wert) / (*Max[Koordinate]-1))) ;
}

float* KartNetz::At (int x,int y,int z)
{ return (&Daten[((z*Max.y()+y)*Max.x()+x)*9]) ;
}

void KartNetz::ParameterAusgabe (ostream& OutStream) const
{ OutStream << AnzahlGitterpunkte << " = " << *Max[0] << " * " << *Max[1] << " * " << *Max[2] << endl
            << "Grenzen:" << endl << Grenzen << endl ;
}

class DreieckInfo 
{ private:
  int DreieckNr ; 
  int EckenNr ; 
  
  public:
  DreieckInfo () : DreieckNr (-1), EckenNr (-1) {} ;
  DreieckInfo (int dreiecknr, int eckennr) : DreieckNr (dreiecknr), EckenNr (eckennr) {} ;
  int GetDreieckNr () const { return (DreieckNr) ; } ;
  int GetEckenNr () const { return (EckenNr) ; } ;
} ;

typedef vector<DreieckInfo> TDreieckInfos ;
class NormalenInfo 
{ private:
  float* Normale ; 
  TDreieckInfos* AngrenzendeDreiecke ;
  
  public:
  NormalenInfo () ;
  NormalenInfo (const NormalenInfo& Original) ;
  NormalenInfo (float* normale) ;
  ~NormalenInfo () ;
  
  NormalenInfo& operator = (const NormalenInfo& Original) ;

  void push_back (const DreieckInfo& dreieckinfo) const
  { 
    (*AngrenzendeDreiecke).push_back (dreieckinfo) ;
  } ;
  void NeuBerechnen (TDreiecke DreieckVektor) const ;
  int Dreiecke () const
  { 
    return ((*AngrenzendeDreiecke).size ()) ; 
  } ;
//  bool operator == (const NormalenInfo& NI) { return (Normale == NI.Normale) ; } ;
//  bool operator < (const NormalenInfo& NI) { return (Normale < NI.Normale) ; } ;
  friend bool operator < (const NormalenInfo NI1, const NormalenInfo NI2) ;
} ;

void NormalenInfo::NeuBerechnen (TDreiecke DreieckVektor) const
{ Matrix<float> NormalenSumme (Null,1,3) ;

  for (TDreieckInfos::iterator Index = (*AngrenzendeDreiecke).begin (); Index != (*AngrenzendeDreiecke).end (); ++Index)
  { 
    NormalenSumme += (DreieckVektor[(*Index).GetDreieckNr ()]).GetNormale() ;
  }
  NormalenSumme /= NormalenSumme.Laenge() ;
  Normale[0] = NormalenSumme.x() ;
  Normale[1] = NormalenSumme.y() ;
  Normale[2] = NormalenSumme.z() ;
}

NormalenInfo::NormalenInfo () : Normale (0)
{ AngrenzendeDreiecke = new TDreieckInfos ; 
} 
NormalenInfo::NormalenInfo (const NormalenInfo& Original) : Normale (Original.Normale)  
{ AngrenzendeDreiecke = new TDreieckInfos (*Original.AngrenzendeDreiecke) ;
} 
NormalenInfo::NormalenInfo (float* normale) : Normale (normale)
{ AngrenzendeDreiecke = new TDreieckInfos ; 
} 
NormalenInfo::~NormalenInfo () 
{ if (AngrenzendeDreiecke) delete AngrenzendeDreiecke ;   
} 

NormalenInfo& NormalenInfo::operator = (const NormalenInfo& Original) 
{ if (this != &Original)
  { delete AngrenzendeDreiecke ;
    Normale = Original.Normale ;
    AngrenzendeDreiecke = new TDreieckInfos (*Original.AngrenzendeDreiecke) ;
  }
  return (*this) ;
}

bool operator < (const NormalenInfo NI1, const NormalenInfo NI2) 
{ return (NI1.Normale < NI2.Normale) ; } ;

typedef less<NormalenInfo> Sortierung ;
typedef set<NormalenInfo,Sortierung> TNormalenInfos ;

void KartNetz::SchreibePov (ostream& OutStream, float AnisotropieLimit, Matrix<float> Magnetfeld, IsooberflaechenArt IOArt)
{ TDreiecke Dreiecke ;
  TPfeile Pfeile ;
  bool* Innerhalb = new bool [AnzahlGitterpunkte] ;
  bool* Oberflaeche = new bool [AnzahlGitterpunkte] ;  
  float* Anisotropie = new float [AnzahlGitterpunkte] ;

  // ber Daten iterieren
  bool* InnerhalbAkt = Innerhalb ;
  float* AnisotropieAkt = Anisotropie ;
  { for (int Z = 0; Z < Max.z(); ++Z)
    { for (int Y = 0; Y < Max.y(); ++Y)
      { for (int X = 0; X < Max.x(); ++X)
        { // Anisotropie berechnen
          Matrix<float> CurMatrix (At(X,Y,Z),3,3) ;

          switch (IOArt)
          { case Aniso:         { *AnisotropieAkt = CurMatrix.Anisotropie () ; break ; } ;
            case AntiSymmAniso: { *AnisotropieAkt = CurMatrix.AntisymmetricAnisotropy () ; break ; } ;
            case FullAniso:     { *AnisotropieAkt = CurMatrix.FullAnisotropy () ; break ; } ;
            case Iso:           { *AnisotropieAkt = abs (CurMatrix.Isotropie ()) ; break ; } ;
            case ICD:           { *AnisotropieAkt = (CurMatrix*Magnetfeld).Laenge () ; break ; } ;
          }

          *InnerhalbAkt++ = *AnisotropieAkt++ >= AnisotropieLimit ;
        }
      }
    }
  }

  cout << "Punkte auf der Isooberflche" << endl ;


  // ber Innerhalb iterieren
  Minimax BildGrenzen ;
  InnerhalbAkt = Innerhalb ;
  bool* OberflaecheAkt = Oberflaeche ;
  {
  for (int Z = 0; Z < Max.z(); ++Z)
  { cout << "Z = " << Z << endl ;

    int minusZ = Max.y()*Max.x() ;
    if (Z == 0) minusZ = 0 ;
    int plusZ = Max.y()*Max.x() ;
    if (Z == Max.z()-1) plusZ = 0 ;

    for (int Y = 0; Y < Max.y(); ++Y)
    { int minusY = Max.x() ;
      if (Y == 0) minusY = 0 ;
      int plusY = Max.x() ;
      if (Y == Max.y()-1) plusY = 0 ;

      for (int X = 0; X < Max.x(); ++X)
      { int minusX = 1 ;
        if (X == 0) minusX = 0 ;
        int plusX = 1 ;
        if (X == Max.x()-1) plusX = 0 ;
   
        // Testen, ob der Punkt innerhalb ist und nicht nur Nachbarn hat, die innerhalb sind
        *OberflaecheAkt++ = (*InnerhalbAkt && !(
           *(InnerhalbAkt+plusX) && *(InnerhalbAkt-minusX) && 
           *(InnerhalbAkt+plusY) && *(InnerhalbAkt-minusY) &&
           *(InnerhalbAkt+plusZ) && *(InnerhalbAkt-minusZ) )) ;
        ++InnerhalbAkt ;

        if (*(OberflaecheAkt-1)) 
        { cout << '#' ;
        } else cout << ((*(InnerhalbAkt-1))? 'o' : '.') ;
      }
      cout << endl ;
    }
    cout << endl ;
  }
  }


  // Koordinaten so verschieben, da der Treppcheneffekt minimiert wird
  int PfeilHinein = 0 ;
  int PfeilHeraus = 0 ;
  float* Orte = new float [3*AnzahlGitterpunkte] ;
  float* PunktNormalen = new float [3*AnzahlGitterpunkte] ;
  int Inkremente [6] ;
  float Step [3] = {Grenzen.DiffX () / Max.x(),Grenzen.DiffY () / Max.y(),Grenzen.DiffZ () / Max.z()} ;

  // ber Oberflaeche iterieren
  AnisotropieAkt = Anisotropie ;
  InnerhalbAkt = Innerhalb ;
  OberflaecheAkt = Oberflaeche ;
  float* OrteAkt = Orte ;
  float* PunktNormalenAkt = PunktNormalen ;
  {
    for (int Z = 0; Z < Max.z(); ++Z)
    { Inkremente[4] = -Max.y()*Max.x() ;
      if (Z == 0) Inkremente[4] = 0 ;
      Inkremente[5] = Max.y()*Max.x() ;
      if (Z == Max.z()-1) Inkremente[5] = 0 ;

      for (int Y = 0; Y < Max.y(); ++Y)
      { Inkremente[2] = -Max.x() ;
        if (Y == 0) Inkremente[2] = 0 ;
        Inkremente[3] = Max.x() ;
        if (Y == Max.y()-1) Inkremente[3] = 0 ;

        for (int X = 0; X < Max.x(); ++X) 
        { if (*OberflaecheAkt++)
          { Inkremente[0] = -1 ;
            if (X == 0) Inkremente[0] = 0 ;
            Inkremente[1] = 1 ;
            if (X == Max.x()-1) Inkremente[1] = 0 ;

            // testen, ob Nachbarpunkt auerhalb liegt, dann die Differenz der Anisotropie merken
            float MinDifferenz [3] = {Gross,Gross,Gross} ;
            int MinInkIndex ;
            { for (int Index = 0; Index < 6; ++Index) if (!(*(InnerhalbAkt + Inkremente[Index])))
              { float AktDiff = *AnisotropieAkt - *(AnisotropieAkt + Inkremente[Index]) ;
                if (AktDiff < abs(MinDifferenz[Index >> 1]))
                { MinDifferenz[Index >> 1] = AktDiff ;
                  if ((Index & 1) == 0) MinDifferenz[Index >> 1] = -AktDiff ;
                  MinInkIndex = Index ;
                }
              }
            }

            // Ort ermitteln
            OrteAkt [0] = RealWert (0,X) ;
            OrteAkt [1] = RealWert (1,Y) ;
            OrteAkt [2] = RealWert (2,Z) ;

            // Differenz der Anisotropie vom Limit im Verhltnis zum geringsten Anisotropieabstand
            // zwischen diesem Punkt und einem Nachbarn
            float DiffZumLimit [3] = 
               { (*AnisotropieAkt - AnisotropieLimit) / MinDifferenz[0] * Step[0],
                 (*AnisotropieAkt - AnisotropieLimit) / MinDifferenz[1] * Step[1],
                 (*AnisotropieAkt - AnisotropieLimit) / MinDifferenz[2] * Step[2] } ;

            // Zhlen, in wievielen Raumkoordinaten eine Verschiebung mglich ist
            int AnzahlRichtungen = 0 ;
            {
              for (int Index = 0; Index < 3; ++Index) if (MinDifferenz[Index] != Gross) ++AnzahlRichtungen ;
            }
            assert (AnzahlRichtungen > 0) ;

            // Verschiebungsvektor berechnen
            float VerschiebungsVektor [3] = {0.0,0.0,0.0} ;
            switch (AnzahlRichtungen)
            { case 1:
              { int Index = MinInkIndex >> 1 ;
                VerschiebungsVektor [Index] = DiffZumLimit [Index] ;
//              if ((MinInkIndex & 1) == 0) VerschiebungsVektor[Index] = -VerschiebungsVektor[Index] ;
                break ;
              }
              case 2:
              { if (MinDifferenz[0] == Gross)
                { float Faktor = DiffZumLimit[1]*DiffZumLimit[2]/(sqr (DiffZumLimit[1]) + sqr (DiffZumLimit[2])) ;
                      VerschiebungsVektor [1] = DiffZumLimit[2] * Faktor ;
                      VerschiebungsVektor [2] = DiffZumLimit[1] * Faktor ;
                } else
                if (MinDifferenz[1] == Gross)
                { float Faktor = DiffZumLimit[0]*DiffZumLimit[2]/(sqr (DiffZumLimit[0]) + sqr (DiffZumLimit[2])) ;
                      VerschiebungsVektor [0] = DiffZumLimit[2] * Faktor ;
                      VerschiebungsVektor [2] = DiffZumLimit[0] * Faktor ;
                } else
                { float Faktor = DiffZumLimit[0]*DiffZumLimit[1]/(sqr (DiffZumLimit[0]) + sqr (DiffZumLimit[1])) ;
                      VerschiebungsVektor [0] = DiffZumLimit[1] * Faktor ;
                      VerschiebungsVektor [1] = DiffZumLimit[0] * Faktor ;
                }        
                break ;
              }
              case 3:
              { VerschiebungsVektor [0] = DiffZumLimit[2]*DiffZumLimit[1] ;
                VerschiebungsVektor [1] = DiffZumLimit[0]*DiffZumLimit[2] ;
                VerschiebungsVektor [2] = DiffZumLimit[1]*DiffZumLimit[0] ;
                float Faktor = DiffZumLimit[0]*DiffZumLimit[1]*DiffZumLimit[2] /
                               (sqr (VerschiebungsVektor[0]) + sqr (VerschiebungsVektor[1]) + sqr (VerschiebungsVektor[2])) ;
                for (int Index = 0; Index < 3; ++Index) VerschiebungsVektor[Index] *= Faktor ;
                break ;
              }
            }

            // Verschiebung durchfhren
            { 
              for (int Index = 0; Index < 3; ++Index) OrteAkt[Index] += VerschiebungsVektor[Index] ;
            }

            // Verschiebungsvektor als Normale merken
            { 
              float Normalisierung = sqrt (sqr (VerschiebungsVektor[0]) + sqr (VerschiebungsVektor[1]) + sqr (VerschiebungsVektor[2])) ;
              for (int Index = 0; Index < 3; ++Index) PunktNormalenAkt[Index] = VerschiebungsVektor[Index] / Normalisierung ;
            }

            // Pfeil generieren
            Matrix<float> CurMatrix (At(X,Y,Z)) ;
            Matrix<float> Origin (&OrteAkt[0],1,3) ;
// so ist es richtig:
            Matrix<float> Pfeilrichtung (CurMatrix*Magnetfeld) ;
// nur Debugging: (Anfang)
//            Matrix<float> Pfeilrichtung (&PunktNormalenAkt[0],1,3) ;
// nur Debugging! (Ende)

            // Mindestgre fr Pfeile: 2e-4, da mit einer Genauigkeit von
            // 1e-4 ausgegeben wird (es sei denn, die Koordinaten bersteigen einen
            // Wert von 10). Falls ein Pfeil 2e-4 lang ist, kann selbst
            // im ungnstigsten Fall durch Rundungsfehler nicht die Pfeillnge auf
            // Null schrumpfen.
            //
            // komischerweise funktioniert die Abfrage quasi berhaupt nicht.
            // Es werden Pfeile durchgelassen, die zu klein sind.
            // Deshalb wird bei der Ausgabe der Pfeil noch einmal geprft.
            //
            // Mglicher Grund: Umrechnung von Bohr in Angstrom ?!?
            if (Pfeilrichtung.Laenge() >= 2e-4)
            {
              Matrix<float> VerschiebungTemp (VerschiebungsVektor,1,3) ;
              if (VerschiebungTemp.SkalarProdukt (Pfeilrichtung) < 0.0)
              { 
                Pfeile.push_back (Pfeil (Origin-Pfeilrichtung,Origin,B_konstant)) ;
                ++PfeilHinein ;
              } else 
              { 
                Pfeile.push_back (Pfeil (Origin,Origin+Pfeilrichtung,A_konstant)) ;
                ++PfeilHeraus ;
              }
            }
            else
            {
      /*      cerr << "Hoppla! Dieser Pfeil ist wohl zu kurz geraten." << endl ;
              cerr << Origin << endl ;
              cerr << Origin+Pfeilrichtung << endl ;
              cerr << Pfeilrichtung << endl ;
              cerr << Pfeilrichtung.Laenge() << endl ;
              // Debug
      */
            }

          // Bildgrenzen ermitteln
            BildGrenzen.CheckGrenzen (0,0,AnisotropieLimit,AnisotropieLimit,OrteAkt[0],OrteAkt[0],OrteAkt[1],OrteAkt[1],OrteAkt[2],OrteAkt[2]) ;
          }

          ++InnerhalbAkt ;
          ++AnisotropieAkt ;
          OrteAkt += 3 ;
          PunktNormalenAkt += 3 ;
        }
      }
    }
  }

  cerr << "Es wurden " << PfeilHinein+PfeilHeraus << " Pfeile generiert." << endl 
       << "Davon zeigen " << PfeilHinein << " in die Isooberflche hinein (" << float (PfeilHinein)/(PfeilHinein+PfeilHeraus)*100 << "%)." << endl ;

  float MittelPfeilLaenge = 0 ;
  { for (TPfeile::iterator Index = Pfeile.begin(); Index != Pfeile.end(); ++Index) 
        MittelPfeilLaenge += (*Index).Laenge () ;
  }
  MittelPfeilLaenge /= (PfeilHinein+PfeilHeraus) ;

  cerr << endl << "Mittelwert der Pfeillnge: " << MittelPfeilLaenge << endl ;

  MittelPfeilLaenge /= float (0.626) ;

  { for (TPfeile::iterator Index = Pfeile.begin(); Index != Pfeile.end(); ++Index) 
        (*Index) /= MittelPfeilLaenge ;
  }

  // Pfeilstatistik erstellen und ausgeben
  float PfeilSumme = 0 ;
  int Pfeilstatistik [12] = {0,0,0,0,0 ,0,0,0,0,0 ,0,0} ;
  {
    for (TPfeile::iterator Index = Pfeile.begin(); Index != Pfeile.end(); ++Index) 
    { 
      float Laenge = (*Index).Laenge () ;
      if (Laenge < 0) ++Pfeilstatistik[0] ;
      else if (Laenge > 1) ++Pfeilstatistik[11] ;
      else ++Pfeilstatistik[int (Laenge*10)+1] ;
      PfeilSumme += Laenge ;
    }
  }
  cerr << "Pfeilstatistik:" << endl ;
  for (int intIndex = 0; intIndex < 12; ++intIndex) cerr << Pfeilstatistik[intIndex] << "  " ;

  cerr << endl << "Dreiecke werden generiert. Bitte warten." << endl ;


  // set mit den Normalen und Dreiecksdaten anlegen
  TNormalenInfos NormalenInfos ;

  // ber Oberflaeche iterieren
  OberflaecheAkt = Oberflaeche ;

  {
    for (int Z = 0; Z < Max.z(); ++Z)
    { for (int Y = 0; Y < Max.y(); ++Y)
      { for (int X = 0; X < Max.x(); ++X) if (*OberflaecheAkt++)
        { // Koordinaten und Anzahl der Nachbarn bestimmen
          Matrix<float> Origin (&Orte[((Z*Max.y()+Y)*Max.x()+X)*3],1,3) ;
          float* OriginNormale (&PunktNormalen[((Z*Max.y()+Y)*Max.x()+X)*3]) ;

          for (int zz = max(Z-1,0); zz < min(Z+2,Max.z()); zz++)
          { for (int yy = max(Y-1,0); yy < min(Y+2,Max.y()); yy++)
            { for (int xx = max(X-1,0); xx < min(X+2,Max.x()); xx++)            
              { // Nur die Punkte beachten, die auch auf der Oberflche sind
                if (((zz != Z) || (yy != Y) || (xx != X)) && (Oberflaeche[(zz*Max.y()+yy)*Max.x()+xx]))
                { Matrix<float> Nachbar1 (&Orte[((zz*Max.y()+yy)*Max.x()+xx)*3],1,3) ;
                  float* Nachbar1Normale (&PunktNormalen[((zz*Max.y()+yy)*Max.x()+xx)*3]) ;

                  for (int zzz = zz; zzz < min(Z+2,Max.z()); zzz++)
                  { for (int yyy = (zzz==zz?yy:max(Y-1,0)); yyy < min(Y+2,Max.y()); yyy++)
                    { for (int xxx = ((zzz==zz)&&(yyy==yy)?xx:max(X-1,0)); xxx < min(X+2,Max.x()); xxx++)
                      { if (((zzz != zz) || (yyy != yy) || (xxx != xx)) && 
                             ((zzz != Z) || (yyy != Y) || (xxx != X)) &&
                            (Oberflaeche[(zzz*Max.y()+yyy)*Max.x()+xxx]))
                        { if ((abs(zzz-zz) < 2) && (abs(yyy-yy) < 2) && (abs(xxx-xx) < 2))
                          { Matrix<float> Nachbar2 (&Orte[((zzz*Max.y()+yyy)*Max.x()+xxx)*3],1,3) ;
                            float* Nachbar2Normale (&PunktNormalen[((zzz*Max.y()+yyy)*Max.x()+xxx)*3]) ;
                          
//                            cout << "Dreieck Nummer: " << Dreiecke.size() +1 << '\015' ; cout.flush () ;
                          
                            TNormalenInfos::iterator Position ;
                            NormalenInfo NI (OriginNormale) ;
                            if ((Position = NormalenInfos.find (NI)) == NormalenInfos.end())
                            {
                              Position = (NormalenInfos.insert (NI)).first ;
                            } ;
                            (*Position).push_back (DreieckInfo (Dreiecke.size(),0)) ;

                            NormalenInfo NI1 (Nachbar1Normale) ;
                            if ((Position = NormalenInfos.find (NI1)) == NormalenInfos.end())
                            {
                              Position = (NormalenInfos.insert (NI1)).first ;
                            } ;
                            (*Position).push_back (DreieckInfo (Dreiecke.size(),1)) ;

                            NormalenInfo NI2 (Nachbar2Normale) ;
                            if ((Position = NormalenInfos.find (NI2)) == NormalenInfos.end())
                            {
                              Position = (NormalenInfos.insert (NI2)).first ;
                            } ;
                            (*Position).push_back (DreieckInfo (Dreiecke.size(),2)) ;

                            Dreiecke.push_back (Dreieck (Origin,Nachbar1,Nachbar2,OriginNormale,Nachbar1Normale,Nachbar2Normale)) ;

/*                             cout << "Dreieck Nummer: " << Dreiecke.size() ;
                            cout << " Normale Nummer: " << NormalenInfos.size () ;
                            cout << " Pos: " << (*Position).Dreiecke () ;
                            cout << " NI: " << NI.Dreiecke () ;
                            cout << " NI1: " << NI1.Dreiecke () ;
                            cout << " NI2: " << NI2.Dreiecke () << endl ;
 */                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
          *(OberflaecheAkt-1) = false ;
        }
      }
    }
  }
  
  // Normalen neu berechnen
  cout << "Neuberechnen der Normalen. Bitte warten." << endl ;

  int ProzentZlr = NormalenInfos.size()/100 ;
  int ProzentZahl = 0 ;
  int NochZuBearbeiten = NormalenInfos.size() - ProzentZlr ;

  { for (TNormalenInfos::iterator NI = NormalenInfos.begin(); NI != NormalenInfos.end(); ++NI)
    { (*NI).NeuBerechnen (Dreiecke) ;
      if (--ProzentZlr == 0)
      { cout << char (13) << ++ProzentZahl << " %" ;
        cout.flush () ;
        ProzentZlr = NochZuBearbeiten/(100-(ProzentZahl==100 ? 99 : ProzentZahl)) ;
        NochZuBearbeiten -= ProzentZlr ;
      }
    }
  }
  cout << endl ;

  cerr << "Povray-input wird geschrieben. Bitte warten." << endl ;

  // Legende schreiben
  OutStream << "#declare Legende=" << endl
            << "union" << endl
            << "{ cylinder { < 0,0.1,0 >,< 0.44096,0.1,0 >,0.0304 pigment { Green }}" << endl
            << "  cone { < 0.44096,0.1,0 >,0.0608,< 0.52918,0.1,0 >,0 pigment { Red }}" << endl
            << "  text" << endl
            << "  { ttf \"timrom.ttf\" \"entspricht " << MittelPfeilLaenge << "\" 1, 0" << endl 
            << "    pigment { color Black }" << endl
            << "    scale 0.33" << endl
            << "    translate < 0.7,0,0 >" << endl
            << "  }" << endl 
            << "}" << endl ;

  OutStream << "#declare Isooberflaeche =" << endl << "mesh" << endl << '{' ;
  
  // Alle Dreiecke ausgeben
  {
    for (TDreiecke::iterator Index3eck = Dreiecke.begin(); Index3eck != Dreiecke.end (); ++Index3eck)
        OutStream << *Index3eck ;
  }

  OutStream << '}' << endl << endl ;

  OutStream << "#declare Pfeile =" << endl << "union" << endl << '{' ;
  
  // Alle Pfeile ausgeben
  { 
    for (TPfeile::iterator Index = Pfeile.begin(); Index != Pfeile.end (); ++Index)
        OutStream << *Index ;
  }

  OutStream << '}' << endl ;

  // Bildgrenzen ausgeben:
  cout << "Bildgrenzen:" << endl << BildGrenzen << endl ;

  // Speicher freigeben
  delete [] Innerhalb ;
  delete [] Oberflaeche ;
  delete [] Anisotropie ;
  delete [] Orte ;
}

/**********************/
/*                    */
/* globale Funktionen */
/*                    */
/**********************/

void Hilfe () 
{ cout << endl 
       << "AICD-smooth_isosurface - Erzeugung eines input-files fuer POV-Ray zur" << endl
       << "                         Visualisierung der Stromdichte" << endl
       << endl
       << "Dieses Programm generiert aus dem Output von AICD-remap ein" << endl
       << "include-file, das die Informationen fuer eine Isooberflaeche" << endl
       << "und fuer die sich auf dieser Oberflaeche befindenden Stromdichte-" << endl
       << "vektoren enthaelt." << endl
       << endl
       << "Die Isooberflaeche wird durch ein Limit und die Art der skalaren" << endl
       << "Groesse, die dargestellt werden soll, bestimmt. Es kann die Iso-" << endl
       << "tropie der Stromdichte, die Anisotropie der Stromdichte und der" << endl
       << "Betrag der Stromdichtevektoren bei einem gegebenen Magnetfeld" << endl
       << "als skalare Groesse verwendet werden." << endl
       << endl 
       << "Aufruf: AICD-smooth_isosurface  limit  Magnetfeld  Isooberflche  input-file(s)" << endl
       << "Output: input-file.inc" << endl
       << "Limit: Grenzwert der Isooberflaeche" << endl
       << "Magnetfeld: Hierfuer muessen drei Werte angegeben werden. Diese" << endl
       << "            werden als X-, Y- und Z-Komponente des Magnetfeldvektors" << endl
       << "            verwendet." << endl
       << "Isooberflche: Drei verschiedene Schluesselwoerter bestimmen, welche" << endl
       << "               skalare Groesse fuer die Isooberflaeche verwendet wird:" << endl
       << "               ICD : Betrag der Stromdichtevektoren" << endl 
       << "               Iso : Isotropie der Stromdichte" << endl 
       << "               Aniso : Anisotropie der Stromdichte" << endl
       << endl 
       << endl ;
} 

/*****************/
/*               */
/* Hauptprogramm */
/*               */
/*****************/

int main (int argc, char* argv [])
{ if (argc < 7) Hilfe () ; else
  for (int ArgIndex = 6; ArgIndex < argc; ++ArgIndex)
  { char* Inputfilename ;

    float Limit ;
    istringstream LimitEinleser (argv[1]) ;
    LimitEinleser >> Limit ;

    Matrix<float> Magnetfeld (Null,1,3) ;
    istringstream MagnetfeldEinleser ((string(argv[2])+' '+string(argv[3])+' '+string(argv[4])).c_str()) ;
    MagnetfeldEinleser >> Magnetfeld ;

    IsooberflaechenArt IOArt ;
    string IOArtString (argv[5]) ;
         if (IOArtString == "ICD")           IOArt = ICD ;
    else if (IOArtString == "Iso")           IOArt = Iso ;
    else if (IOArtString == "Aniso")         IOArt = Aniso ; 
    else if (IOArtString == "AntiSymmAniso") IOArt = AntiSymmAniso ; 
    else if (IOArtString == "FullAniso")     IOArt = FullAniso ; 
    else cerr << IOArtString << " ist keine zulssige Angabe fr den Parameter" << endl
              << "Isooberflche. Es wird Isooberflche = Aniso angenommen." << endl ;

    // Parameter einlesen    
    Inputfilename = argv[ArgIndex] ;
    string Outputfilename = string(Inputfilename)+".inc";

    cout << endl << ArgIndex-5 << ". Input-file: " << Inputfilename << endl ;

    // Einlesen der Daten aus dem input-file
#ifdef LINUX    
    ifstream Inputfile (Inputfilename) ;
#else
#ifdef IRIX
    ifstream Inputfile (Inputfilename) ;
#else
    ifstream Inputfile (Inputfilename,ios::binary) ;
#endif
#endif
    KartNetz globalesNetz (Inputfile) ;
    globalesNetz.ParameterAusgabe (cout) ;

    // PovRay-Datei schreiben
   
    ofstream PovDatei (Outputfilename.c_str()) ;
    PovDatei << "#declare Magnetfeldvektor=" << endl 
             << "union" << endl 
             << "{ cylinder { < " 
             << -5*Magnetfeld.x() << "," << -5*Magnetfeld.y() << "," << -5*Magnetfeld.z() 
             << " >,< "
             << 2.5*Magnetfeld.x() << "," << 2.5*Magnetfeld.y() << "," << 2.5*Magnetfeld.z() 
             << " >,0.15 }" << endl
             << "  cone { < " 
             << 2.5*Magnetfeld.x() << "," << 2.5*Magnetfeld.y() << "," << 2.5*Magnetfeld.z() 
             << " >,0.3,< "
             << 5*Magnetfeld.x() << "," << 5*Magnetfeld.y() << "," << 5*Magnetfeld.z() 
             << " >,0.0 }" << endl
             << "}" << endl ;
    PovDatei << "#declare Aufrufzeile=" << endl
             << "text" << endl
             << "{ ttf \"timrom.ttf\" \"" ;
    for (int Index = 0; Index < argc; ++Index) PovDatei << argv[Index] << ' ' ;
    PovDatei << "\" 1, 0" << endl
             << "  pigment { color Black }" << endl
             << "}" << endl ;
    globalesNetz.SchreibePov (PovDatei,Limit,Magnetfeld,IOArt) ;
  }
  return 0;
}







